# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import report_import
report_import("zef.core.graph_additions.command_preparation")

from .common import *

##############################################
# * Preparation dispatch
#--------------------------------------------

def dispatch_preparation(cmd, gs, context):
    return (cmd,gs,context) | match_rules[[
        *preparation_rules,
        (Any, not_implemented_error["Unknown cmd type for preparation"]),
    ]] | collect


# TODO: This should be possible to process even at interpretation phase if we
# know that a new object is to be created. Hence, this code should be split off,
# into a part that handles the "no object on graph" and a part that handles an
# existing object. They can use the same code for final application.
def prepare_obj_notation(cmd, gs, context):
    gen_id_state = context["gen_id_state"]
    from .wish_tagging import ensure_tag, Taggable

    cmds = []

    cmd, me, gen_id_state = ensure_tag(cmd, gen_id_state)

    z_on_graph = None
    need_to_create = False
    if isinstance(me, EntityRef):
        raise Exception("Shouldn't get here anymore")
        # cmds += [me]
        # names = RAET_get_names(et)
        # if len(names) > 0:
        #     cmds += [PleaseAlias(ids=(force_as_id(me),) + names)]
        # if me in gs:
        #     z_on_graph = gs | get[me] | collect
    elif isinstance(me, WishID):
        if rae_type(cmd) is None:
            pass
        elif isinstance(rae_type(cmd), RT):
            need_to_create = False
        else:
            need_to_create = True
    elif isinstance(me, EternalUID):
        if isinstance(rae_type(cmd), RT):
            need_to_create = False
        else:
            z_on_graph = find_rae_in_target(me, gs)
            if z_on_graph is None:
                need_to_create = True
    else:
        raise Exception(f"Object notation got a weird value for its own ID: {me}")

    if need_to_create:
        # TODO: Because the "pure ET with a name" should actually be an atom
        # (the ET.x[y] is a legacy style), this might have to change into
        # something more directly the pure ET creation, i.e. a PleaseInstantiate
        # itself directly.
        if isinstance(rae_type(cmd), PureET | PureAET):
            # TODO: Probably need to handle multiple names here at some point
            # assert len(get_names(cmd)) == 1
            cmds += [rae_type(cmd)[me]]
        elif isinstance(rae_type(cmd), PureRT):
            raise Exception("Shouldn't get here anymore")
        else:
            raise NotImplementedError(f"TODO can't create something without a known type: {cmd}")


    from ..atom import _get_fields
    for k, v in _get_fields(cmd).items():
        rels_exact = False
        rt = RT(k)
        if isinstance(v, PrimitiveValue | Taggable | UserWishID):
            v = {v}

        if isinstance(v, set):
            # First go through and figure out the minimal set of changes.
            #
            # to_create = make a new relation and possibly a new AE (if it is a value)
            # to_terminate = termiante an old relation
            # to_assign = reuse an existing relation + AE and assign a new value
            # to_keep = leave this relation alone
            if z_on_graph is None:
                to_create = v
                to_terminate = []
                to_assign = []
                to_keep = []
            else:
                to_create = []
                to_assign = []
                to_keep = []

                existing_rels = z_on_graph | out_rels[rt] | collect
                existing_items = {}
                for rel in existing_rels:
                    t = target(rel)
                    if isinstance(t, AttributeEntity):
                        t_val = value(t)
                    else:
                        t_val = None
                    existing_items[rel] = (t, t_val)
                
                assign_candidates = {}
                existing_free = set(existing_rels)
                for item in v:
                    if isinstance(item, UserWishID):
                        to_create += [item]
                        continue
                    this_assign_candidates = []
                    for existing_rel in existing_free:
                        existing_target,existing_val = existing_items[existing_rel]
                        if isinstance(item, AtomClass):
                            if discard_frame(item) == discard_frame(existing_target):
                                to_keep += [existing_rel]
                                existing_free.remove(existing_rel)
                                break
                        else:
                            if item == existing_val:
                                to_keep += [existing_rel]
                                existing_free.remove(existing_rel)
                                break
                            if can_assign_to(item, existing_target):
                                this_assign_candidates += [existing_rel]
                    else:
                        if isinstance(item, AtomClass):
                            to_create += [item]
                        else:
                            assign_candidates[item] = this_assign_candidates

                for to_decide,candidates in assign_candidates.items():
                    candidates = set(candidates) & existing_free 
                    if len(candidates) == 0:
                        to_create += [to_decide]
                    else:
                        candidate = list(candidates)[0]
                        to_assign += [(to_decide, candidate)]
                        existing_free.remove(candidate)

                to_terminate = list(existing_free)
                        

            exact_ids = []

            for rel in to_keep:
                exact_ids += [origin_uid(rel)]

            for rel in to_terminate:
                cmds += [PleaseTerminate(target=origin_uid(rel))]

            for val,rel in to_assign:
                exact_ids += [origin_uid(rel)]
                cmds += [PleaseAssign(target=origin_uid(target(rel)), value=Val(val))]

            for item in to_create:
                item,item_id,gen_id_state = ensure_tag(item, gen_id_state)
                cmds += [item]
                id_rel,gen_id_state = gen_internal_id(gen_id_state)
                cmds += [PleaseInstantiate(
                    atom=dict(rt=rt, source=me, target=item_id),
                    internal_ids=[id_rel]
                )]
                exact_ids += [id_rel]

            cmds += [PleaseBeSource(target=force_as_id(me), rel_ids=exact_ids, exact=True, rt=rt)]

        # elif type(v) in {list, tuple}:
        #     list_id = gen_id()
        #     cmds.append( (me, RT(to_pascal_case(k)), ET.ZEF_List[list_id]) )

        #     # generate ids for each relation, that we can inter-connect them
        #     list_ids = [gen_id() for _ in range(len(v))]
        #     cmds.extend(list_ids 
        #             | sliding[2] 
        #             | map[lambda p: (Z[p[0]], RT.ZEF_NextElement, Z[p[1]])]
        #             | collect
        #         )
        #     for el, edge_id in zip(v, list_ids):
        #         sub_obj_instrs = expand_helper(el, gen_id)
        #         cmds.append( (ET.ZEF_List[list_id], RT.ZEF_ListElement[edge_id], sub_obj_instrs[0]) )
        #         cmds.extend(sub_obj_instrs[1:])
        # else:
        #     cmds.append( (me, RT(to_pascal_case(k)), v) )
        else:
            raise NotImplementedError(f"TODO: obj notation for {v}")

    context = context | insert["gen_id_state"][gen_id_state] | collect
    return [], cmds, context

def prepare_please_run(cmd, gs, context):
    todo = []
    cmds = []
    if isinstance(cmd.action, LazyValue):
        # Execute op one after another
        cur = cmd.action.initial_val
        ops = list(cmd.action.el_ops)
        for op in ops:
            rt,abs = peel(op)[0]
            if RT[rt] == RT.OutRel:
                names = abs | filter[Variable] | collect
                op = ZefOp(((rt, abs | filter[~Variable] | collect),))
            else:
                names = []

            cur = op(cur)
            if len(names) > 0:
                cmds += [PleaseAlias(ids=[origin_uid(cur)] + names)]

        if isinstance(cur, SetFieldAction):
            todo += [PleaseRun(action=cur)]
        else:
            raise NotImplementedError(f"TODO: PleaseRun result {cur}")
    else:
        raise NotImplementedError("TODO: PleaseRun")

    return cmds, todo, context

def prepare_relations(cmd, gs, context):
    # Reltaions getting to this point are pure, but they need to indicate their
    # implicit behaviour on the source/target nodes to effectively "claim
    # territory" that other relation-creating commands (like the EVI) need to
    # cooperate with.
    out_cmds = []
    gen_id_state = context["gen_id_state"]

    names = cmd | get["internals_ids"][[]] | collect
    if len(names) == 0:
        id,gen_id_state = gen_internal_id(gen_id_state)
        names = [id]
        cmd = cmd._get_type()(cmd._value | insert["internal_ids"][names])
    id = names[0]

    out_cmds += [cmd]
    out_cmds += [PleaseBeSource(target=force_as_id(src), rel_id=id, exact=False, rt=cmd.atom["rt"])]
    out_cmds += [PleaseBeTarget(target=force_as_id(trg), rel_id=id, exact=False, rt=cmd.atom["rt"])]

    context = context | insert["gen_id_state"][gen_id_state] | collect
    return out_cmds, [], context

def prepare_pass_through(cmd, gs, context):
    return [cmd], [], context

def prepare_interpret(cmd, gs, context):
    # raise NotImplementedError("TODO fallback to interpret")
    from .wish_interpretation import generate_level2_commands, default_interpretation_rules
    cmds,todo,context = (cmd,context) | match_rules[[
        *default_interpretation_rules,
        (Any, not_implemented_error["Unknown cmd type for interpretation during preparation"]),
    ]] | collect
    # All outputs are todo from the point of view of preparation vs interpration.
    return [], cmds + todo, context

preparation_rules = [
    (AtomClass, prepare_obj_notation),
    (PleaseRun, prepare_please_run),
    (PleaseInstantiate & Is[get["atom"] | is_a[PleaseInstantiateRelation]], prepare_relations),
    (PleaseCommandLevel1, prepare_pass_through),
    # Fallback to interpretation rules here but make them pass back into this list always (i.e. redirect the lists)
    (GraphWishInput, prepare_interpret),
]
    


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import report_import
report_import("zef.core.rae_type_definitions")

from .VT import *
from .VT import insert_VT, make_VT

Relation = insert_VT("Relation", RelationRef | RelationConcrete | RelationAtom | RelationFlatRef)
AttributeEntity = insert_VT("AttributeEntity", AttributeEntityRef | AttributeEntityConcrete | AttributeEntityAtom | AttributeEntityFlatRef)
Entity = insert_VT("Entity", EntityRef | EntityConcrete | EntityAtom | EntityFlatRef)
RAE = insert_VT('RAE', Entity | AttributeEntity | Relation)

TX = insert_VT("TX", TXNodeConcrete | TXNodeRef | TXNodeAtom | TXNodeFlatRef)
Root = insert_VT("Root", RootConcrete | RootRef | RootAtom | RootFlatRef)

def Atom_is_a(x, typ):
    return isinstance(x, RAE | TX | Root)
def Atom_ctor(self, *args, **kwargs):
    return AtomClass(*args, **kwargs)
    
Atom = make_VT("Atom",
               is_a_func=Atom_is_a,
               constructor_func=Atom_ctor,
               pass_self=True)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *
import zef

from zef.core.graph_additions.types import *
from zef.core.graph_additions.low_level import *
from zef.core.graph_additions.wish_translation2 import *
from zef.core.graph_additions.wish_interpretation import *


class MyTestCase(unittest.TestCase):

    def test_level1(self):
        g = Graph()
        g2 = Graph()

        cmd = PleaseInstantiate({"atom": ET.Machine,
                                 "internal_ids": [V.asdf]})
        cmds = [cmd]
        concrete_cmd_list = Level1CommandInfo({"cmds": cmds,
                                               "gs": now(g),
                                               "resolved_variables": {}})
        gs_updated,receipt = perform_level1_commands(concrete_cmd_list, False)

        self.assertSetEqual(set(receipt.keys()), {V.asdf})
        self.assertIsInstance(receipt[V.asdf], ET.Machine)

        cmd = PleaseInstantiate({"atom": ET.Machine,
                                 "internal_ids": [V.two],
                                 "origin_uid": origin_uid(receipt[V.asdf])})
        cmds = [cmd]
        concrete_cmd_list = Level1CommandInfo({"cmds": cmds,
                                               "gs": now(g2),
                                               "resolved_variables": {}})
        gs_updated,receipt2 = perform_level1_commands(concrete_cmd_list, False)
        self.assertEqual(origin_uid(receipt[V.asdf]), origin_uid(receipt2[V.two]))

    def test_generate_level1(self):
        # lvl1_rules = default_translation_rules

        g = Graph()
        z = ET.Machine | g | run
        g2 = Graph()
        
        cmds = [PleaseInstantiate({"atom": ET.Machine,
                                   "internal_ids": [V.one]}),
                PleaseInstantiate({"atom": ET.Machine,
                                   "internal_ids": [V.two]}),
                PleaseInstantiate({"atom": ET.Machine,
                                   "origin_uid": origin_uid(z),
                                   "internal_ids": [V.three]}),
                PleaseInstantiate({"atom": ET.Machine,
                                   "origin_uid": origin_uid(z),
                                   "internal_ids": [V.four]})]

        output_cmds = generate_level1_commands(cmds, now(g))#, lvl1_rules)
        self.assertEqual(output_cmds.gs, now(g))
        self.assertSetEqual(set(output_cmds.cmds),
                         {PleaseInstantiate({"atom": ET.Machine,
                                             "internal_ids": [V.one]}),
                          PleaseInstantiate({"atom": ET.Machine,
                                             "internal_ids": [V.two]})})
        output_cmds2 = generate_level1_commands(cmds, now(g2))#, lvl1_rules)
        self.assertSetEqual(set(output_cmds2.cmds),
                         {PleaseInstantiate({"atom": ET.Machine,
                                             "internal_ids": [V.one]}),
                          PleaseInstantiate({"atom": ET.Machine,
                                             "internal_ids": [V.two]}),
                          PleaseInstantiate({"atom": ET.Machine,
                                             "origin_uid": origin_uid(z)})})

        gs_updated,receipt = perform_level1_commands(output_cmds2, False)
        self.assertEqual(set(receipt.keys()), {V.one, V.two, V.three, V.four})
        self.assertEqual(receipt[V.three], receipt[V.four])
        self.assertNotEqual(receipt[V.one], receipt[V.two])
        self.assertNotEqual(receipt[V.one], receipt[V.three])
        self.assertNotEqual(receipt[V.two], receipt[V.three])

        cmds = [PleaseTerminate({"target": origin_uid(z)}),
                PleaseInstantiate({"atom": ET.Machine,
                                   "internal_ids": [V.thisframe]}),
                PleaseTerminate({"target": V.thisframe})]
        with self.assertRaisesRegex(Exception, "Can't have two commands for the same name"):
            output_cmds3 = generate_level1_commands(cmds, now(g))#, lvl1_rules)

    def test_generate_level2(self):
        # lvl1_rules = default_translation_rules
        lvl2_rules = default_interpretation_rules

        g = Graph()
        z = ET.Machine | g | run
        g2 = Graph()
        
        info = generate_level2_commands([ET.Machine[V.one]], lvl2_rules)
        temp = PleaseInstantiate({"atom": ET.Machine, "internal_ids": [V.one]})
        self.assertEqual(info["cmds"], [PleaseInstantiate({"atom": ET.Machine,
                                                   "internal_ids": (V.one,)})])
        output_cmds = generate_level1_commands(info["cmds"], now(g))#, lvl1_rules)
        receipt = perform_level1_commands(output_cmds, False)


        info = generate_level2_commands([
            ET.Machine[V.one],
            AET.Int[V.other],
            (V.one, RT.Something[V.two], V.other),
        ], lvl2_rules)

        # TODO:
        # self.assertEqual(info, {"cmds": [
        #     PleaseInstantiate({"atom": ET.Machine,
        #                        "internal_ids": [V.one]}),
        #     PleaseInstantiate({"atom": AET.Int,
        #                        "internal_ids": [V.other]}),
        #     PleaseInstantiate({"atom": {"rt": RT.Something,
        #                                 "source": V.one,
        #                                 "target": V.other,
        #                                 "combine_source": False,
        #                                 "combine_target": False},
        #                        "internal_ids": [V.two]}),
        # ]})

        info = generate_level2_commands([
            ET.Machine[V.one](something=5),
            (V.one, RT.extra, V.one),
        ], lvl2_rules)
        cmds = info["cmds"]

        # TODO:
        # self.assertEqual(cmds[0], PleaseInstantiate({"atom": ET.Machine,
        #                                              "internal_ids": [V.one]}))
        # print(cmds)
        # print("==============")
        # print("==============")
        # print("==============")
        # print("==============")
        # print(cmds[1])
        # self.assertEqual(cmds[1].atom, AET.Int)
        # self.assertEqual([cmds[2].target], cmds[1].internal_ids)
        # self.assertEqual(cmds[2].value, Val(5))
        # self.assertEqual(cmds[3], PleaseInstantiate({"atom": {"rt": RT.something,
        #                                                       "source": V.one,
        #                                                       "target": cmds[1].internal_ids[0],
        #                                                       "combine_source": False,
        #                                                       "combine_target": False,}}))
        # self.assertEqual(cmds[4], PleaseInstantiate({"atom": {"rt": RT.extra,
        #                                                       "source": V.one,
        #                                                       "target": V.one,
        #                                                       "combine_source": False,
        #                                                       "combine_target": False}}))
        # self.assertEqual(len(cmds), 5)

        lvl1_cmds = generate_level1_commands(cmds, now(g))#, lvl1_rules)
        receipt = perform_level1_commands(lvl1_cmds, False)

    def test_obj_notation(self):
        g = Graph()

        z = ET.Location | g | run
        ae = AET.Int | g | run
        # loc = ET.Location(z)
        loc = ET.Location("㏈-51ec2f5e07f6063c1eb21d36ef2ccf19")

        inputs = [
            ET.Person[V.bob](name="Bob", favourite_food={"cheese", "pizza"}),
            ET.Person[V.jane](name="Jane", favourite_food={"pasta", "steak"}),

            (V.bob, RT.MarriedTo(V.married, meta="meta-test"), V.jane),

            V.married | set_field[RT.Date][Time("1999-01-01")],
            (V.married, RT.Location, [z, loc]),

            # loc.name << "seafront",
            # loc.country << "antartica",
            loc(people_visited={V.bob,V.jane}),

            ae | assign[42],
        ]


        from zef.core.graph_additions.types import GraphWishInput
        for input in inputs:
            self.assertIsInstance(input, GraphWishInput)

        gs_updated,receipt = inputs | transact[g] | run

        married_rel = receipt[V.married]
        z_married_rel = gs_updated | get[married_rel] | collect
        self.assertEqual(z_married_rel | F.meta | collect, "meta-test")
        self.assertEqual(value(now(ae)), 42)



        z_no_prior = ET.Machine | g | run
        z_prior = ET.Machine(something=42) | g | run
        # TODO: Things with sets here
        [
            Atom(z_no_prior)(something=43),
            Atom(z_prior)(something=43),
        ] | g | run
        self.assertEqual(z_no_prior | now | F.something | collect, 43)
        self.assertEqual(z_prior | now | F.something | collect, 43)


    def test_backwards_compatibility(self):
        g = Graph()

        z = ET.Location | g | run
        z2 = {ET.Location: {RT.Details: 0}} | g | run

        inputs = [
            {ET.Person["person"]: {RT.FirstName["a"]: "Joe", RT.LastName["b"]: "Bloggs"}},
            {Z["person"]: {RT.Location: "Antartica"}},
            (Z["person"], RT.Weird, Z["a"]),
            {Z["a"]: {RT.MetaData: "stuff"}},

            (Z["b"], RT.MetaData["meta"], 42),
            (Z["meta"], RT.Something["meta2"], 42),
            
            z | set_field[RT.Something][42],
            # z2 | Out[RT.Details] | assign[1],
        ]

        gs_updated,receipt = inputs | transact[g] | run


    def test_all_encode(self):
        # This is meant to follow the list that is in `shorthand.py`. It mostly
        # just checks to see if the cmds run, rather than their returns.
        g = Graph()

        z = ET.Machine(something=42) | g | run
        a,b,c = (z, RT.Something, 42) | g | run
        a,_,cs = (z, RT.Relation, [ET.Machine, ET.Machine]) | g | run
        delegate_of(ET.Machine) | g | run
        c | assign[43] | g | run
        z | g | run
        z | terminate | g | run
        {ET.Machine: {RT.Something: 42}} | g | run
        Val(42) | g | run

    def test_all_interpret(self):
        # This is meant to follow the list that is in `wish_interpretation.py`.
        # It mostly just checks to see if the cmds run, rather than their
        # returns.
        g = Graph()

        z = ET.Machine | g | run
        z2 = ET.Machine | g | run
        z_ae = AET.Int | g | run

        gs_updated,r = [
            ET.Machine(V.x, something=42),
            (V.x, RT.Something(V.relation, meta=1), 42),
            (z, RT.Relation, [ET.Machine, ET.Machine]),
            delegate_of(ET.Machine),
            z_ae | assign[43],
            z,
            z2 | terminate,
            {ET.Machine: {RT.Something: 42}},
            Val(42),
            ET.Machine[V.random],
            AET.Int[V.random2],
        ] | transact[g] | run

        # The collected variants
        z2 = ET.Machine | g | run
        z_ae = AET.Int | g | run

        gs_updated,r = [
            z_ae | assign[43] | collect,
            z2 | terminate | collect,
        ] | transact[g] | run

        with self.assertRaises(Exception):
            [
                ET.Machine(V.x),
                ET.Other(V.x),
            ] | transact[g] | run

        with self.assertRaises(Exception):
            [
                (ET.Machine, RT.Something(V.x), 42),
                ET.Other(V.x),
            ] | transact[g] | run

        with self.assertRaises(Exception):
            z = ET.Machine | g | run
            g2 = Graph()
            [
                z,
                z | terminate,
            ] | transact[g2] | run

        with self.assertRaises(Exception):
            [
                ET.Machine[V.something],
                V.something | terminate,
            ] | transact[g] | run
            

        


if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import report_import
report_import("zef.core.graph_additions.types")

__all__ = [
    "PleaseInstantiateEntity",
    "PleaseInstantiateRelation",
    "PleaseInstantiateAttributeEntity",
    "PleaseInstantiateValueNode",
    "PleaseInstantiateDelegate",
    "PleaseInstantiate",
    "PleaseTerminate",
    "PleaseAssign",
    "PleaseTag",
    "PleaseCommandLevel1",
    "Level1CommandInfo",

    "PleaseMustLive",
    "PleaseBeSource",
    "PleaseBeTarget",
    "PleaseAlias",

    "ResolvedVariableDict",

    "PleaseRun",
    "PleaseCommandLevel2",
    "Level2CommandInfo",

    "PureET",
    "PureRT",
    "PureAET",

    "RelationTriple",
    "PrimitiveValue",
    "WrappedValue",
    "AETWithValue",
    "OldStyleDict",
    "OldStyleRelationTriple",

    "GraphWishValue",
    "GraphWishInputSimple",
    "GraphWishInput",
    "GraphWishReceipt",

    "WishID",
    "WishIDInternal",
    "InternalIDs",
    "FlatRefUID",
    "Variable",
    "AllIDs",
    "UserWishID",
    "ExtraUserAllowedIDs",
    "NamedZ",
    "NamedAny",
]


# Implementation questions - should ValueTypes or UserValueTypes be preferred for these kinds of things?
# Pros:
# - ValueTypes don't include a tag, e.g. PleaseTerminate, so are easier to
#   create future versions that extend previous types. e.g. a new key in the
#   dictionary.
# - UserValueTypes do include a tag, so are easier to avoid accidental incorrect
#   use in older versions rather than erroring. Extra code is required for
#   converting between versions but it can be made explicit.
# Cons:
# - ValueTypes need custom code to describe their objects (i.e. being a
#   dictionary is actually several lines to describe). This could be alleviated
#   with factory code to generate ValueTypes.
# - See the opposites of the Pros.

from ..VT import *
from ..VT import make_VT
from .. import _ops
from ..atom import FlatRefUID, FlatGraphPlaceholder

def non_zefop_alias(vt, name):
    from ..VT import ValueType_
    vt2 = ValueType_(type_name=vt._d['type_name'], absorbed=vt._d['absorbed'])
    vt2._d['alias'] = name
    return vt2
_alias = non_zefop_alias

# IDs are allowed for many aspects of linking between parts of a wish. We want
# to allow variables, but need to include strings for backwards compoatibility
# for a version or two.
# WishID = String | Variable
Variable = _alias(SymbolicExpression & Is[_ops.get_field["root_node"] | _ops.equals[None]],
                 "Variable")
    
# Internal wish ids as used to reference items, but are themselves not returned in the receipt.
WishIDInternal = UserValueType("WishIDInternal", String)

InternalIDs = WishIDInternal | FlatRefUID

WishID = _alias(Variable | InternalIDs,
               "WishID")

# This is to allow us potentially switching up the intention of Val in the main
# user namespace
WrappedValue = Val

AllIDs = WishID | EternalUID | DelegateRef | WrappedValue

NamedAny = ValueType[Any] & Is[_ops.absorbed | _ops.length | _ops.greater_than[0]]
def delayed_check_namedz(x):
    if not isinstance(x, ZExpression):
        return False
    if _ops.rae_type(x.root_node) != ET.GetItem:
        return False
    if x.root_node.arg1 != ET.Z:
        return False
    return True
NamedZ = Is[delayed_check_namedz]

ExtraUserAllowedIDs = NamedZ | NamedAny

UserWishID = WishID | ExtraUserAllowedIDs

PrimitiveValue = _alias(PyInt | String | PyFloat | PyBool | Time | Enum | QuantityInt | QuantityFloat,
                       "PrimitiveValue")

# This is to work around deprecation issues
PureET = ValueType & ET
PureRT = ValueType & RT
PureAET = ValueType & AET


######################################
# * Level 1 commands
#------------------------------------
# Level 1 commands - precise commands that are direct actions on a graph slice.
# These can be generated from a set of levl 2 commands and a graph slice and
# should then be *guaranteeed* to be applicable, in constrast to level 2
# commands.

PleaseInstantiateEntity = PureET
PleaseInstantiateRelation = Pattern[{"rt": PureRT,
                                     "source": AllIDs,
                                     "target": AllIDs}]
PleaseInstantiateAttributeEntity = PureAET
PleaseInstantiateValueNode = WrappedValue
PleaseInstantiateDelegate = DelegateRef

PleaseInstantiateAtom = _alias(PleaseInstantiateEntity
                                           | PleaseInstantiateRelation
                                           | PleaseInstantiateAttributeEntity
                                           | PleaseInstantiateValueNode
                                           | PleaseInstantiateDelegate,
                                   "PleaseInstantiateAtom")
PleaseInstantiate = UserValueType("PleaseInstantiate",
                                  Dict,
                                  Pattern[{"atom": PleaseInstantiateAtom,
                                           Optional["origin_uid"]: EternalUID,
                                           Optional["internal_ids"]: List[WishID]}]
                                  # An exception to the rule, is that value nodes and delegates cannot have origin_uids
                                  & ~Pattern[{"atom": PleaseInstantiateDelegate | PleaseInstantiateValueNode,
                                              "origin_uid": Any}]
                                  )

PleaseTerminate = UserValueType("PleaseTerminate",
                                Dict,
                                Pattern[{
                                    "target": AllIDs,
                                    # Note: including internal_ids here makes
                                    # some of the logic trickier. Going to
                                    # exclude this possibility for the moment,
                                    # but can introduce it later on if needed.
                                    #"internal_ids": Optional[List[WishID]]
                                }])

PleaseAssign = UserValueType("PleaseAssign",
                              Dict,
                              Pattern[{"target": AllIDs,
                                       "value": WrappedValue}])

PleaseTag = UserValueType("PleaseTag",
                          Dict,
                          Pattern[{"target": RAERef | AllIDs,
                                   Optional["internal_ids"]: List[WishID]}])

# Only for internal use to resolve potential invalid graph wishes.
PleaseMustLive = UserValueType("PleaseMustLive",
                               Dict,
                               Pattern[{"target": AllIDs}])
# Only for internal use to handle set_field etc
PleaseBeSource = UserValueType("PleaseBeSource",
                               Dict,
                               Pattern[{"target": AllIDs,
                                        "rel_ids": List[AllIDs],
                                        "exact": Bool,
                                        "rt": RT,
                                        }])
PleaseBeTarget = UserValueType("PleaseBeTarget",
                               Dict,
                               Pattern[{"target": AllIDs,
                                        "rel_ids": [AllIDs],
                                        "exact": Bool,
                                        "rt": RT,
                                        }])
PleaseAlias = UserValueType("PleaseAlias",
                            Dict,
                            Pattern[{"ids": List[AllIDs]}])

PleaseCommandLevel1 = _alias(PleaseInstantiate
                             | PleaseAssign
                             | PleaseTerminate
                             | PleaseTag
                             | PleaseMustLive
                             | PleaseBeSource
                             | PleaseBeTarget
                             | PleaseAlias,
                             "PleaseCommandLevel1")




# An import delay
def are_commands_ordered(info):
    from .command_ordering import are_commands_ordered
    return are_commands_ordered(info["cmds"], info["gs"])
OrderedCommands = Is[are_commands_ordered]
# When prepared, a list of commands must be in the context of a graph slice
ResolvedVariableDict = Dict[Variable][EternalUID | Variable]
Level1CommandInfo = UserValueType(
    "Level1CommandInfo",
    Dict,
    Pattern[{"gs": GraphSlice | FlatGraph,
             "cmds": List[PleaseCommandLevel1],
             # Resolved variables are those that are identified at translation
             # time, in contrast to instantiations that will happen at action
             # time.
             "resolved_variables": ResolvedVariableDict,
             }] & OrderedCommands)


##############################
# * Level 2 commands
#----------------------------
# Level 2 commands - declaration of what to be done, although intent could be
# for a particular graph slice, could be attempted to be applied on a later
# graph slice.

PleaseRun = UserValueType("PleaseRun",
                          Dict,
                          Pattern[{"action": LazyValue | ZefOp | ZExpression}])

PleaseCommandLevel2 = _alias(PleaseCommandLevel1 | PleaseRun | AtomClass,
                                 "PleaseCommandLevel2")
Level2CommandInfo = UserValueType("Level2CommandInfo",
                                  Dict,
                                  Pattern[{"cmds": List[PleaseCommandLevel2],
                                           Optional["custom"]: Any}])

##############################
# * Main interface input
#----------------------------


# These are the external graph wish interface possibilities. This is the most
# flexible, which is converted to level 2 commands in order to transmit to the
# transactor.

# RelationTriple = _alias(Tuple[UserWishID | Atom | PrimitiveValue,
#                                   PureRT,
#                                   UserWishID | Atom | PrimitiveValue],
#                             "RelationTriple")
# TODO: A RelationTriple should recursively allow anything that is a GraphWishInput for its source and target
def is_relation_triple(x):
    if not isinstance(x, List):
        return False
    if not len(x) == 3:
        return False
    # I'm uncomfortable with EternalUID appearing in here - it is not enough
    # information on its own, and must be accompanied by an instantiate
    # elsewhere. But it does appear a lot as part of the iteration
    Item = GraphWishInput | UserWishID | EternalUID
    if not isinstance(x[0], Item):
        return False
    if not isinstance(x[1], PureRT | RelationAtom):
        return False
    if not isinstance(x[2], Item):
        return False
    return True
RelationTriple = _alias(Is[is_relation_triple], "RelationTriple")
def is_relation_triple_OS(x):
    if not isinstance(x, List):
        return False
    Item = GraphWishInput | UserWishID
    if len(x) == 3:
        return (isinstance(x[0], Item | List[Item])
                and isinstance(x[1], PureRT)
                and isinstance(x[2], Item | List[Item]))
    elif len(x) == 2:
        if isinstance(x[0], Item) and isinstance(x[1], List[Tuple[PureRT, Item]]):
            return True
        return False
    else:
        return False
OldStyleRelationTriple = _alias(Is[is_relation_triple_OS], "OldStyleRelationTriple")
GraphWishValue = _alias(PrimitiveValue | WrappedValue, "GraphWishValue")

# This is a declaration of *always* instantiating a new AE with the given AET
# and initializing it with a value. It primarily exists for internal use at the
# lvl2 phase, although there's no reason a user couldn't also write this.
AETWithValue = UserValueType("AETWithValue",
                             Dict,
                             Pattern[{"aet": PureAET,
                                      "value": Any, # TODO: This should be a parametric type related to the AET
                                      Optional["internal_ids"]: List[WishID]}])


# Backwards compatibility
# Any is used here instead of GraphWishInputSimple. It would be nice to have the
# recursive definition but that makes for ugly hacks to implement it.
OldStyleDict = Dict[UserWishID | PureET | RAE][Dict[PureRT][Any]] & Is[_ops.length | _ops.equals[1]]

# "Simple" inputs are those that can be interpreted almost (excepting EUID references) directly into lvl1 commands
GraphWishInputSimple = _alias(
    LazyValue
    | SymbolicExpression
    | RelationTriple
    | OldStyleRelationTriple
    | GraphWishValue
    | AETWithValue
    | FlatGraph
    | PureET
    | PureAET
    | RAE
    | TX
    | Root
    | Delegate
    | OldStyleDict,
    "GraphWishInputSimple"
)

GraphWishInput = _alias(PleaseCommandLevel2 | GraphWishInputSimple,
                        "GraphWishInput")

GraphWishReceipt = Dict[AllIDs][AtomRef]


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import report_import
report_import("zef.core.graph_additions.transact")

# This is the interface for the user

# Notes:
# - rules for interpretation (input -> lvl2 cmds) are acted upon in transact.
# - rules for transaltion (lvl2 -> lvl1) are acted upon in the effect.
# - lvl1 has no rules. It is the lowest-lvl commands that map 1-to-1 to the
#   operations applied to a graph.

# To better handle sending graph wishes over the wire, while zef functions are
# not serializable, we will use a string "default" for the translation rules.


from .common import *


@func
def transact_dispatch(input, onto, *others):
    # if isinstance(onto, (Graph, GraphRef)):
    if True:
        return transact_graph(input, onto, *others)
    elif isinstance(onto, FlatGraph):
        from ..op_implementations import fg_insert_imp
        return fg_insert_imp(onto, input)
    else:
        raise Exception(f"Don't understand kind of object to transact onto: {onto}")

@func
def transact_graph(input: GraphWishInput | List[GraphWishInput], g, interpretation_rules=None, translation_rules="default", post_transact_rule=None):
    # assert isinstance(input, GraphWishInput | List[GraphWishInput])
    # Some helpful output before we have proper "explain" options
    if not isinstance(input, GraphWishInput | List[GraphWishInput]):
        if isinstance(input, List):
            for x in input:
                if not isinstance(x, GraphWishInput):
                    raise TypeError(f"The following is not a valid GraphWishInput: {x}")
        raise TypeError(f"The following is not a valid GraphWishInput: {input}")
    if isinstance(input, GraphWishInput):
        input = [input]

    if interpretation_rules is None:
        from .wish_interpretation import default_interpretation_rules
        interpretation_rules = default_interpretation_rules
        
    from .wish_interpretation import generate_level2_commands
    cmds = generate_level2_commands(input, interpretation_rules)

    if isinstance(g, Graph | GraphRef):
        target_ref = GraphRef(g)
    elif isinstance(g, FlatGraph):
        target_ref = g
    else:
        raise Exception("Unknown target for a transaction")

    from ..fx import FX
    return {
        "type": FX.Graph.Transact,
        "target_glike": target_ref,
        "level2_commands": cmds,
        "translation_rules": translation_rules,
        "post_transact_rule": post_transact_rule,
    }
    



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *
import zef

class MyTestCase(unittest.TestCase):

    def test_atom(self):
        from zef.core.atom import _get_ref_pointer
        g = Graph()

        atom = ET.Machine | g | run
        z = _get_ref_pointer(atom)

        self.assertIsInstance(atom, Atom)
        self.assertEqual(_get_ref_pointer(to_ezefref(atom)), to_ezefref(z))
        self.assertEqual(rae_type(atom), rae_type(z))
        self.assertEqual(preceding_events(atom), preceding_events(z))
        self.assertEqual(uid(atom), uid(z))
        self.assertEqual(exists_at(atom, now(g)), exists_at(z, now(g)))
        self.assertEqual(frame(atom), frame(z))
        self.assertEqual(discard_frame(atom), discard_frame(z))

        atom2 = AET.Int | g | run
        atom2 | assign[42] | g | run
        atom2 = now(atom2)
        z2 = _get_ref_pointer(atom2)
        self.assertEqual(value(atom2), value(z2))

        from zef.core.VT.rae_types import RAET_get_token
        rel = zef.pyzef.main.instantiate(to_ezefref(z), RAET_get_token(RT.Something), to_ezefref(z2), g)

        atom_rel = Atom(rel)

        self.assertEqual(_get_ref_pointer(now(atom)), now(_get_ref_pointer(atom)))

        z = now(z)
        atom = now(atom)
        z2 = now(z2)
        atom2 = now(atom2)

        self.assertEqual(source(atom_rel), atom)
        self.assertEqual(source(rel), z)
        self.assertEqual(target(atom_rel), atom2)
        self.assertEqual(target(rel), z2)

        self.assertEqual(Out(atom, RT.Something), atom2)
        self.assertEqual(out_rel(atom, RT.Something), atom_rel)
        self.assertEqual(Out(z, RT.Something), z2)
        self.assertEqual(out_rel(z, RT.Something), rel)

        self.assertEqual(In(z2, RT.Something), z)
        self.assertEqual(in_rel(z2, RT.Something), rel)

        self.assertEqual(has_out(atom, RT.Something), has_out(z, RT.Something))
        self.assertEqual(has_in(atom, RT.Something), has_in(z, RT.Something))

        self.assertEqual(_get_ref_pointer(select_by_field([atom], RT.Something, 42)), select_by_field([z], RT.Something, 42))


if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import report_import
report_import("zef.core.flat_graph")

from operator import ne
from dataclasses import dataclass
from .VT import *
from .VT import make_VT


class FlatGraph_:
    """
    Internal data representation:
    1) self.blobs: a tuple of blobs
    2) self.key_dict: a dictionary mapping keys to indices in self.blobs

    Each element of self.blobs is a tuple of the form 
    (
        index: Int, 
        blob_type: e.g. ET.Foo / BT.VALUE_NODE / RT.Bar / AET.Int,
        edge_list: a list of blob indexes (integers). Positive for outgoing, negative for incoming
        origin_uid (optional)
    )
    """
    def __init__(self, *args):
        from ._ops import insert, collect
        if args == ():
            self.key_dict = {}
            self.blobs = ()
        elif len(args) == 1 and isinstance(args[0], list):
            new_fg = FlatGraph_()
            new_fg = new_fg | insert[args[0]] | collect
            self.key_dict = new_fg.key_dict
            self.blobs = new_fg.blobs
        elif len(args) == 1 and isinstance(args[0], FlatRef_):
            self.key_dict =  args[0].fg.key_dict
            self.blobs = args[0].fg.blobs
        else:
            raise NotImplementedError("FlatGraph with args")

    def __repr__(self):
        kdict = "\n".join([f"({k}=>{v})" for k,v in self.key_dict.items()])
        blobs = "\n".join([str(e) for e in self.blobs])
        return f'FlatGraph(\n{kdict}\n-------\n{blobs}\n)'
    
    def __or__(self, other):
        from .VT import LazyValue
        return LazyValue(self) | other

    def __ror__(self, other):
        from ._ops import insert
        return self | insert[other] 

    def __getitem__(self, key):
        from ._ops import get
        return get(self, key)

    def __contains__(self, key):
        if isinstance(key, Val):
            from ._ops import value_hash
            return value_hash(key.arg) in self.key_dict
        else:
            return key in self.key_dict

FlatGraph = make_VT("FlatGraph", pytype=FlatGraph_)


class FlatRef_:
    def __init__(self, fg, idx):
        self.fg = fg
        self.idx = idx
    def __repr__(self):
        return f'<FlatRef #{abs(self.idx)} {repr(self.fg.blobs[self.idx][1])}>'
    
    def __or__(self, other):
        from .VT import LazyValue
        return LazyValue(self) | other

    def __gt__(self, other):
        from .VT import LazyValue
        return LazyValue(self) > other

    def __lt__(self, other):
        from .VT import LazyValue
        return LazyValue(self) < other

    def __lshift__(self, other):
        from .VT import LazyValue
        return LazyValue(self) << other
    
    def __rshift__(self, other):
        from .VT import LazyValue
        return LazyValue(self) >> other

FlatRef = make_VT("FlatRef", pytype=FlatRef_)

def FlatRef_rae_type(fr):
    return fr.fg.blobs[fr.idx][1]
def FlatRef_value(fr):
    if FlatRef_rae_type(fr) != "BT.ValueNode":
        raise Exception("Not a value node")
    return fr.fg.blobs[fr.idx][1]

class FlatRefs_:
    def __init__(self, fg, idxs):
        self.fg = fg
        self.idxs = idxs
    def __repr__(self):
        newline = '\n'
        return f"""<FlatRefs len={len(self.idxs)}> [
{["    " + repr(FlatRef_(self.fg, i)) for i in self.idxs] | join[newline] | collect}
]"""
    
    def __or__(self, other):
        from .VT import LazyValue
        return LazyValue(self) | other

    def __iter__(self):
        return (FlatRef_(self.fg, i) for i in self.idxs)

    def __len__(self):
        return len(self.idxs)

    def __gt__(self, other):
        from .VT import LazyValue
        return LazyValue(self) > other

    def __lt__(self, other):
        from .VT import LazyValue
        return LazyValue(self) < other

    def __lshift__(self, other):
        from .VT import LazyValue
        return LazyValue(self) << other
    
    def __rshift__(self, other):
        from .VT import LazyValue
        return LazyValue(self) >> other

FlatRefs = make_VT("FlatRefs", pytype=FlatRefs_)



# Danny added to avoid implementation specific coding
def FlatRef_rae_type(fr):
    return fr.fg.blobs[fr.idx][1]

def FlatRef_maybe_uid(fr):
    return fr.fg.blobs[fr.idx][-1]






fg_registry = {}
def register_flatgraph(fg):
    from .VT.value_type import hash_frozen
    h = hash_frozen(fg)
    fg_registry[h] = fg
    return h

def lookup_flatgraph(h):
    return fg_registry.get(h, None)


from .VT import make_VT, insert_VT

def RelationFlatRef_is_a(x, typ):
    from ._ops import rae_type, is_a
    return isinstance(x, FlatRef & Is[rae_type | is_a[RT]])
RelationFlatRef = make_VT("RelationFlatRef", is_a_func=RelationFlatRef_is_a)
def EntityFlatRef_is_a(x, typ):
    from ._ops import rae_type, is_a
    return isinstance(x, FlatRef & Is[rae_type | is_a[ET]])
EntityFlatRef = make_VT("EntityFlatRef", is_a_func=EntityFlatRef_is_a)
def AttributeEntityFlatRef_is_a(x, typ):
    from ._ops import rae_type, is_a
    return isinstance(x, FlatRef & Is[rae_type | is_a[AET]])
AttributeEntityFlatRef = make_VT("AttributeEntityFlatRef", is_a_func=AttributeEntityFlatRef_is_a)
def TXNodeFlatRef_is_a(x, typ):
    from ._ops import abstract_type, equals
    return isinstance(x, FlatRef & Is[abstract_type | equals[BT.TX_EVENT_NODE]])
TXNodeFlatRef = make_VT("TXNodeFlatRef", is_a_func=TXNodeFlatRef_is_a)
def RootFlatRef_is_a(x, typ):
    from ._ops import abstract_type, equals
    return isinstance(x, FlatRef & Is[abstract_type | equals[BT.ROOT_NODE]])
RootFlatRef = make_VT("RootFlatRef", is_a_func=RootFlatRef_is_a)



RAEFlatRef = insert_VT('RAEFlatRef', EntityFlatRef | AttributeEntityFlatRef | RelationFlatRef)



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import report_import
report_import("zef.core.graph_additions.transact")

from .common import *

@func
def encode(input: GraphWishInput | List[GraphWishInput], g):
    gen_id_state = generate_initial_state("encode")
    cmds, template, gen_id_state = encode_cmd(input, gen_id_state)

    #return out_cmds, out_template

    t_effect = cmds | transact[g] | collect

    return (t_effect
            | insert["unpacking_template"][template]
            | insert["keep_internal_ids"][True]
            | collect)

def dispatch_ror_graph(g, x):
    return x | encode[g]

from ...pyzef import main
main.Graph.__ror__ = dispatch_ror_graph


def encode_cmd(obj, gen_id_state: GenIDState):
    from .wish_tagging import tagging_rules as default_rules

    all_ensure_tag_types = Union[tuple(default_rules | map[first] | collect)]

    encode_rules = [
        (RelationTriple, encode_relation_triple),
        (OldStyleRelationTriple, encode_OS_relation_triple),
        (List, encode_list),
        (PleaseTerminate, encode_terminate),

        # Fallback to ensure tag
        (all_ensure_tag_types, encode_ensure_tag_fallback[default_rules]),
    ]

    return (obj, gen_id_state) | match_rules[[
        *encode_rules,
        (Any, not_implemented_error["Don't know how to encode object for shorthand syntax"]),
    ]] | collect

def make_name_for_rt(rt, gen_id_state):
    names = names_of_raet(rt)
    if len(names) == 0:
        me,gen_id_state = gen_internal_id(gen_id_state)
        rt = rt[me]
    else:
        me = names[0]

    return rt,me,gen_id_state

def encode_relation_triple(obj, gen_id_state):
    s,rt,t = obj

    s_cmds, s_tag, gen_id_state = encode_cmd(s, gen_id_state)
    t_cmds, t_tag, gen_id_state = encode_cmd(t, gen_id_state)

    rt,me,gen_id_state = make_name_for_rt(rt, gen_id_state)

    cmds = s_cmds + t_cmds + [(s_tag, rt, t_tag)]
    tags = [s_tag, me, t_tag]
    return cmds, tags, gen_id_state

def encode_OS_relation_triple(obj, gen_id_state):
    if len(obj) == 3:
        s,rt,t = obj

        s_cmds, s_tags, gen_id_state = encode_cmd(s, gen_id_state)
        t_cmds, t_tags, gen_id_state = encode_cmd(t, gen_id_state)

        cmds = s_cmds + t_cmds
        tags = (s_tags, None, t_tags)

        if not isinstance(s_tags, List):
            s_tags = [s_tags]
        if not isinstance(t_tags, List):
            t_tags = [t_tags]
        for s_tag in s_tags:
            for t_tag in t_tags:
                cmds += [(s_tag, rt, t_tag)]

        return cmds, tags, gen_id_state
    elif len(obj) == 2:
        s,rels = obj
        s_cmds, s_tag, gen_id_state = encode_cmd(s, gen_id_state)

        cmds = s_cmds

        rel_names = []
        for rt,t in rels:
            rt,rt_name,gen_id_state = make_name_for_rt(rt, gen_id_state)
            t_cmds, t_tag, gen_id_state = encode_cmd(t, gen_id_state)
            rel_names += [(rt_name, t_tag)]
            cmds += t_cmds
            cmds += [(s_tag, rt, t_tag)]

        tags = (s_tag, rel_names)
            
        return cmds, tags, gen_id_state
    else:
        raise Exception("Shouldn't get here")

def encode_list(obj, gen_id_state):
    out_cmds = []
    out_tags = []
    for item in obj:
        this_cmds,this_tag,gen_id_state = encode_cmd(item, gen_id_state)
        out_cmds += this_cmds
        out_tags += [this_tag]
    return out_cmds, out_tags, gen_id_state

def encode_terminate(obj, gen_id_state):
    return [obj], force_as_id(obj.target), gen_id_state

@func
def encode_ensure_tag_fallback(obj, tagging_rules, gen_id_state):
    obj,id,gen_id_state = (obj, gen_id_state) | match_rules[[
        *tagging_rules,
        (Any, not_implemented_error["Don't know how to ensure a tag for object"]),
    ]] | collect

    return [obj], id, gen_id_state
    


def unpack_receipt(template, receipt, glike_frame):
    if isinstance(glike_frame, GraphSlice):
        return unpack_receipt_graphslice(template, receipt, glike_frame)
    elif isinstance(glike_frame, FlatGraph):
        return unpack_receipt_flatgraph(template, receipt, glike_frame)
    else:
        raise Exception(f"Unexpected frame-like: {glike_frame}")

def unpack_receipt_flatgraph(template, receipt, fg):
    if isinstance(template, List):
        return tuple((unpack_receipt_flatgraph(el, receipt, fg) for el in template))
    if isinstance(template, WishID):
        if isinstance(template, Variable) and isinstance(template.name, OriginallyUserID):
            template = template.name.obj
        # This should probably be an atom already in the receipt
        # return Atom(receipt[template])
        return fg[template]
    if isinstance(template, EternalUID):
        return fg[template]
    if isinstance(template, Nil):
        return None
    if isinstance(template, DelegateRef):
        return fg[template]
    if isinstance(template, WrappedValue):
        return fg[template]
    raise Exception(f"Should not get here - unknown type in unpacking template: {template}")

def unpack_receipt_graphslice(template, receipt, gs):
    if isinstance(template, List):
        return tuple((unpack_receipt(el, receipt, gs) for el in template))
    if isinstance(template, WishID):
        if isinstance(template, Variable) and isinstance(template.name, OriginallyUserID):
            template = template.name.obj
        # This should probably be an atom already in the receipt
        return Atom(receipt[template])
    if isinstance(template, EternalUID):
        from ..graph_slice import get_instance_rae
        out = get_instance_rae(template, gs, allow_tombstone=True)
        assert out is not None, f"Out is None! Coming from {template}"
        return Atom(out)
    if isinstance(template, Nil):
        return None
    if isinstance(template, DelegateRef):
        out = to_delegate(template, gs)
        assert out is not None
        return out
    if isinstance(template, WrappedValue):
        val = internals.val_as_serialized_if_necessary(template)
        out = Graph(gs).get_value_node(val)
        assert out is not None
        return out | in_frame[gs] | collect
    raise Exception(f"Should not get here - unknown type in unpacking template: {template}")


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .common import *

from .wish_tagging import ensure_tag

def evaluate_chain(val, el_ops, gen_id_state):
    state = dict(obj=val,
                 # ops_itr=iter(el_ops),
                 focused_field=None,
                 emitted_cmds=[],
                 gen_id_state=gen_id_state)

    final_state = list(el_ops) | reduce[apply_op][state] | collect

    return final_state

@func
def apply_op(state, op):
    @func
    def op_kind(op):
        return RT[peel(op)[0][0]]
        
    return (op,state) | match_ops[
        (RT.Field, apply_field),
        (RT.Assign, apply_assign),
        (RT.SetField, apply_set_field),
        (RT.Terminate, apply_terminate),
    ] | collect

@func
def match_ops(input, rules):
    op,state = input
    op_rt = RT[peel(op)[0][0]]

    for rule in rules:
        if op_rt == rule[0]:
            return rule[1](op,state)

    raise Exception(f"Shouldn't get here: no rule found for op {op_rt}")
        

    

def apply_field(op, state):
    rt = single(absorbed(op))

    assert state["focused_field"] is None
    return state | insert["focused_field"][rt] | collect
    
def apply_assign(op, state):
    val = single(absorbed(op))

    gen_id_state = state["gen_id_state"]
    new_cmds = []

    if isinstance(state["obj"], AtomClass):
        # assert isinstance(state["focused_field"], RT)
        if state["focused_field"] is None:
            assert isinstance(state["obj"], AttributeEntityAtom)
            new_obj,me,gen_id_state = ensure_tag(state["obj"], gen_id_state)
            # Make sure that we don't include the frame with the atom now, since we
            # are assigning a value to it.
            from .._ops import discard_frame
            new_obj = discard_frame(new_obj)
            new_cmds += [PleaseAssign(target=me, value=Val(val))]
        else:
            new_obj = state["obj"](**{token_name(state["focused_field"]) : val })

        return (state
                | merge[{
                    "focused_field": None,
                    "obj": new_obj,
                    "gen_id_state": gen_id_state,
                    "emitted_cmds": state["emitted_cmds"] + new_cmds,
                }]
                | collect)
    else:
        raise NotImplementedError("TODO: non-atoms in assign")

def apply_set_field(op, state):
    set_target = state["obj"]

    if not isinstance(set_target, AtomClass):
        set_target = Atom(set_target)

    rt,val = absorbed(op)

    new_obj = set_target(**{token_name(rt): val})

    return state | insert["obj"][new_obj] | collect

def apply_terminate(op, state):
    if len(absorbed(op)) > 0:
        raise Exception(f"Terminate should take no additional arguments: {absorbed(op)}.")
    obj = state["obj"]
    if not isinstance(obj, AtomClass):
        raise Exception("Terminate can't understand anything other than an Atom")
    from ..atom import _get_fields
    assert len(_get_fields(obj)) == 0
    id = force_as_id(get_most_authorative_id(obj))
        
    new_obj = PleaseTerminate(target=id)
    return (state
            | insert["obj"][new_obj]
            | collect)




def cast_symbolic_expression_as_lazyvalue(se):
    assert isinstance(se.root_node, ET.Or)

    as_list = extract_binary_tree(se.root_node)

    lv = LazyValue(as_list[0])
    return lv | to_pipeline(as_list[1:])

def extract_binary_tree(fr):
    # Treat this like a binary tree and we have to extract the ordered list
    if isinstance(fr, ET.Or):
        left = extract_binary_tree(fr | Out[RT.Arg1] | collect)
        right = extract_binary_tree(fr | Out[RT.Arg2] | collect)
        return left + right
    elif rae_type(fr) == BT.VALUE_NODE:
        return [value(fr)]
    else:
        raise NotImplementedError(f"TODO: {fr}")



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import report_import
report_import("zef.core.graph_additions.low_level")

from ..VT import *
from .._ops import *
from ..internals import Transaction
from ... import pyzef
from ..VT.rae_types import RAET_get_token

from .common import *

def perform_level1_commands(command_struct: Level1CommandInfo, keep_internal_ids: Bool) -> GraphWishReceipt:
    assert isinstance(command_struct, Level1CommandInfo)

    if len(command_struct.cmds) == 0:
        # There is nothing to do! Let's just return early with the old graph slice
        gs_updated = command_struct.gs
        receipt = {}
        return gs_updated, receipt


    # Run through each command

    # For each command:
    # - resolve required ids
    # - action command
    # - insert into tracking dict the generated id

    gs = command_struct.gs
    g = Graph(gs)

    # The receipt is for items that are returned to the wisher
    receipt: GraphWishReceipt = {}
    # The mapping is a way for us to keep track of items created internally
    internal_mapping: Dict[InternalIDs | WrappedValue | DelegateRef][ZefRef] = {}

    # Optimised euid lookup, instead of looking up using full lineage system all the time
    euid_mapping: Dict[EternalUID][ZefRef] = {}
    def find_euid(euid: EternalUID):
        if euid in euid_mapping:
            return euid_mapping[euid]

        zr = most_recent_rae_on_graph(euid, g)
        euid_mapping[euid] = zr
        return zr

    def find_id(id: AtomRef | AllIDs) -> ZefRef:
        # print("find_id", id)
        if isinstance(id, EternalUID):
            return find_euid(id)
        elif isinstance(id, AtomRef):
            return find_euid(origin_uid(id))
        elif isinstance(id, InternalIDs):
            return internal_mapping[id]
        elif isinstance(id, WrappedValue | DelegateRef):
            if id in internal_mapping:
                return internal_mapping[id]
            # It must already be on the graph
            return now(g) | get[id] | collect
        else:
            assert isinstance(id, Variable)
            return find_euid(origin_uid(receipt[id]))

    def record_id(id: AllIDs, z: ZefRef):
        # print("record_id", id, z)
        if isinstance(id, InternalIDs | WrappedValue | DelegateRef):
            internal_mapping[id] = z
        elif isinstance(id, Variable):
            receipt[id] = Atom(z)
        else:
            raise Exception("Shouldn't get here")


    with Transaction(g) as ctx:
        # Note: we might have a transaction open around us so check both now and
        # the previous graph slices.
        if gs != now(g) and gs != now(g) | time_travel[-1] | collect:
            raise Exception("Can't perform level 1 commands onto a different graph slice from which they were constructed: {now(g)=}, {gs=}")

        for cmd in command_struct.cmds:
            # print("Doing cmd", cmd)
            if isinstance(cmd, PleaseInstantiate):
                # If is a brand-new item (no origin_uid) then create
                # Otherwise, merge in, referencing the original via lineage

                # z will be the object created/merged
                z = None
                if "origin_uid" in cmd:
                    assert isinstance(cmd.origin_uid, EternalUID)
                    assert cmd.origin_uid not in gs

                    if cmd.origin_uid in g:
                        # In this branch, the foreign RAE already exists and we
                        # need to link it back up. This could have happened with
                        # an instance that was created and terminated.
                        raise NotImplementedError("TODO: reviving a terminated instance")
                    else:
                        if isinstance(cmd["atom"], PleaseInstantiateEntity):
                            z = internals.merge_entity_(g, RAET_get_token(cmd.atom), cmd.origin_uid.blob_uid, cmd.origin_uid.graph_uid)
                            z = now(z)
                        elif isinstance(cmd["atom"], PleaseInstantiateAttributeEntity):
                            z = internals.merge_atomic_entity_(g, RAET_get_token(cmd.atom), cmd.origin_uid.blob_uid, cmd.origin_uid.graph_uid)
                            z = now(z)
                        elif isinstance(cmd["atom"], PleaseInstantiateRelation):
                            z_src = find_id(cmd["atom"]["source"])
                            z_trg = find_id(cmd["atom"]["target"])
                            z = internals.merge_relation_(g, RAET_get_token(cmd.atom["rt"]),
                                                          to_ezefref(z_src), to_ezefref(z_trg),
                                                          cmd.origin_uid.blob_uid, cmd.origin_uid.graph_uid)
                            z = now(z)
                        else:
                            raise NotImplementedError(f"TODO cmd.atom for merge: {cmd['atom']}")

                    euid_mapping[cmd.origin_uid] = z
                else:
                    if isinstance(cmd.atom, PleaseInstantiateEntity | PleaseInstantiateAttributeEntity):
                        raet = RAET_get_token(cmd.atom)
                        z = pyzef.main.instantiate(RAET_get_token(cmd.atom), g)
                    elif isinstance(cmd.atom, PleaseInstantiateRelation):
                        z_source = find_id(cmd.atom["source"])
                        z_target = find_id(cmd.atom["target"])
                        z = pyzef.main.instantiate(z_source, RAET_get_token(cmd.atom["rt"]), z_target, g)
                    elif isinstance(cmd.atom, PleaseInstantiateDelegate):
                        z = to_delegate(cmd.atom, now(g), True)
                        # Special case for recording an id
                        record_id(cmd.atom, z)
                    elif isinstance(cmd.atom, PleaseInstantiateValueNode):
                        val = internals.val_as_serialized_if_necessary(cmd.atom)
                        z = pyzef.main.instantiate_value_node(val, Graph(gs))
                        # Special case for recording an id
                        record_id(cmd.atom, z)
                    else:
                        raise NotImplementedError("TODO cmd.atom")

                # TODO: This should really be a "to ref" which would just be a
                # normal ZefRef in the future. For now, we'll do it this way,
                # but have the final processing of the receipt to turn these
                # into ZefRefs, just like the old graph delta.
                if "internal_ids" in cmd:
                    # print("About to record", z)
                    for id in cmd.internal_ids:
                        record_id(id, z)
            elif isinstance(cmd, PleaseAssign):
                z = find_id(cmd.target)
                internals.assign_value_imp(z, cmd.value.arg)
            elif isinstance(cmd, PleaseTerminate):
                z = find_id(cmd.target)
                pyzef.zefops.terminate(z)
            else:
                raise NotImplementedError(f"TODO cmd: {cmd}")

    # We keep track of the exact GraphSlice resulting from this transaction
    gs_updated = GraphSlice(ctx)

    # The receipt also gains the predetermined variables that were able to be
    # resovled before reaching this function.
    for k,alias in command_struct.resolved_variables.items():
        assert isinstance(k, Variable)
        record_id(k, find_id(alias))

    if keep_internal_ids:
        receipt.update(internal_mapping)

    # We undo any non-variable user ids based upon their included value
    receipt = maybe_unwrap_variables_in_receipt(receipt)
     
    return gs_updated, receipt



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..VT import RT
from .implementation_typing_functions import * 
from .yo import yo_implementation, yo_type_info
from .graphviz import graphviz_imp, graphviz_tp
from . to_flatgraph import to_flatgraph_imp
from .func import unpack_implementation, unpack_type_info
from .. import internals
from ..fx.fx_runtime import fx_runtime
from .data_wrangling import infer_types, deduplicate, identify_entities

_op_to_functions = {
        internals.RT.ApplyFunctions: (apply_functions_imp, None),
        internals.RT.Map:            (map_implementation, map_type_info),
        internals.RT.MapCat:         (map_cat_imp, map_cat_tp),
        internals.RT.Reduce:         (reduce_implementation, reduce_type_info),
        internals.RT.Scan:           (scan_implementation, scan_type_info),
        internals.RT.GroupBy:        (group_by_implementation, group_by_type_info),
        internals.RT.Group:          (group_imp, None),
        internals.RT.Transpose:      (transpose_imp, transpose_tp),
        internals.RT.Frequencies:    (frequencies_imp, frequencies_tp),
        internals.RT.Iterate:        (iterate_implementation, iterate_type_info),
        internals.RT.Identity:       (identity_implementation, identity_type_info),
        internals.RT.Length:         (length_implementation, length_type_info),
        internals.RT.Count:          (count_imp, None),
        internals.RT.Take :          (take_implementation, take_type_info),
        internals.RT.TakeWhile:      (take_while_imp, take_while_tp),
        internals.RT.TakeWhilePair:  (take_while_pair_imp, take_while_pair_tp),
        internals.RT.TakeUntil:      (take_until_imp, take_until_tp),
        internals.RT.SkipWhile:      (skip_while_imp, skip_while_tp),
        internals.RT.SkipUntil:      (skip_until_imp, None),
        internals.RT.Skip :          (skip_imp, skip_tp),
        internals.RT.Nth:            (nth_implementation, nth_type_info),
        internals.RT.Filter:         (filter_implementation, filter_type_info),
        internals.RT.SelectKeys:     (select_keys_imp, None),
        internals.RT.Modulo:         (modulo_imp, None),
        internals.RT.SelectByField:  (select_by_field_imp, select_by_field_tp),
        internals.RT.Without:        (without_imp, without_tp),
        internals.RT.First:          (first_imp, None),
        internals.RT.Second:         (second_imp, second_tp),
        internals.RT.Last:           (last_imp, last_tp),
        internals.RT.Single:         (single_imp, single_tp),
        internals.RT.SingleOr:       (single_or_imp, single_or_tp),
        internals.RT.Zip:            (zip_imp, zip_tp),
        internals.RT.Concat:         (concat_implementation, concat_type_info),
        internals.RT.Prepend:        (prepend_imp, prepend_tp),
        internals.RT.Append:         (append_imp, append_tp),
        internals.RT.Interleave:     (interleave_imp, interleave_tp),
        internals.RT.InterleaveLongest: (interleave_longest_imp, interleave_longest_tp),
        internals.RT.Sort:            (sort_implementation, sort_type_info),
        internals.RT.Now:             (now_implementation, now_type_info),
        internals.RT.Events:          (events_imp, events_tp),
        internals.RT.PrecedingEvents: (preceding_events_imp, None),
        internals.RT.ToDelegate:      (to_delegate_implementation, to_delegate_type_info),
        internals.RT.DelegateOf:      (delegate_of_implementation, delegate_of_type_info),
        internals.RT.IsBlueprintAtom: (is_blueprint_atom_imp, None),
        
        internals.RT.In:              (In_imp, None),
        internals.RT.Ins:             (Ins_imp, None),
        internals.RT.Out:             (Out_imp, None),
        internals.RT.Outs:            (Outs_imp, None),
        internals.RT.InsAndOuts:      (ins_and_outs_imp, None),
        internals.RT.InRel:           (in_rel_imp, None),
        internals.RT.InRels:          (in_rels_imp, None),
        internals.RT.OutRel:          (out_rel_imp, None),
        internals.RT.OutRels:         (out_rels_imp, None),
        
        internals.RT.Source:          (source_implementation, source_type_info),
        internals.RT.Target:          (target_implementation, target_type_info),
        internals.RT.Value:           (value_implementation, None),
        internals.RT.Time:            (time_implementation, time_type_info),
        internals.RT.TimeSlice:       (time_slice_implementation, time_slice_type_info),
        internals.RT.NextTX:          (next_tx_imp, next_tx_tp),
        internals.RT.PreviousTX:      (previous_tx_imp, previous_tx_tp),
        internals.RT.InstantiationTx: (instantiation_tx_implementation, instantiation_tx_type_info),                              # TODO: retire
        internals.RT.TerminationTx:   (termination_tx_implementation, termination_tx_type_info),                            # TODO: retire
        internals.RT.Uid:             (uid_implementation, uid_type_info),
        internals.RT.Frame:           (frame_imp, frame_tp),
        internals.RT.DiscardFrame:    (discard_frame_imp, None),
        internals.RT.BaseUid:         (base_uid_implementation, base_uid_type_info),
        internals.RT.OriginUid:       (origin_uid_imp, origin_uid_tp),
        internals.RT.ExistsAt:        (exists_at_implementation, exists_at_type_info),
        internals.RT.AwareOf:         (aware_of_implementation, None),
        internals.RT.IsZefRefPromotable:(is_zefref_promotable_implementation, is_zefref_promotable_type_info),
        internals.RT.ToFrame:         (to_frame_imp, to_frame_tp),
        internals.RT.ToGraphSlice:    (to_graph_slice_imp, to_graph_slice_tp),
        internals.RT.ToTx:            (to_tx_imp, to_tx_tp),
        internals.RT.TimeTravel:      (time_travel_imp, time_travel_tp),
        internals.RT.ToEZefRef:       (to_ezefref_implementation, to_ezefref_type_info),
        internals.RT.Terminate:       (terminate_implementation, terminate_type_info),
        internals.RT.Assign:          (assign_imp, assign_tp),
        internals.RT.ET:              (ET_implementation, ET_type_info),
        internals.RT.RT:              (RT_implementation, RT_type_info),
        internals.RT.AET:             (AET_implementation, AET_type_info),
        internals.RT.BT:              (BT_implementation, BT_type_info),
        internals.RT.SetField:        (set_field_implementation, set_field_type_info),
        internals.RT.Assert:          (assert_implementation, assert_type_info),
        internals.RT.HasOut:          (hasout_implementation, hasout_type_info),
        internals.RT.HasIn:           (hasin_implementation, hasin_type_info),
        internals.RT.Run:             (fx_runtime, None),
        internals.RT.Tap:             (tap_imp, tap_tp),
        internals.RT.Push:            (push_imp, push_tp),
        internals.RT.IsA:             (is_a_implementation, is_a_type_info),
        internals.RT.IsRepresentedAs: (is_represented_as_implementation, is_represented_as_type_info),
        internals.RT.RepresentationType: (representation_type_imp, None),
        internals.RT.HasRelation:     (has_relation_implementation, has_relation_type_info),
        internals.RT.Relation:        (relation_implementation, relation_type_info),
        internals.RT.Relations:       (relations_implementation, relations_type_info),
        internals.RT.Chunk:           (chunk_imp, chunk_tp),
        internals.RT.Sliding:         (sliding_imp, sliding_tp),
        internals.RT.Stride:          (stride_imp, stride_tp),
        internals.RT.Insert:          (insert_imp, insert_tp),
        internals.RT.InsertInto:      (insert_into_imp, None),
        internals.RT.ReverseArgs:     (reverse_args_imp, None),
        internals.RT.Remove:          (remove_imp, remove_tp),
        internals.RT.Get:             (get_imp, get_tp),
        internals.RT.GetField:        (get_field_imp, get_field_tp),
        internals.RT.Enumerate:       (enumerate_imp, enumerate_tp),
        internals.RT.Items:           (items_imp, items_tp),
        internals.RT.Values:          (values_imp, values_tp),
        internals.RT.Keys:            (keys_imp, keys_tp),
        internals.RT.Reverse:         (reverse_imp, reverse_tp),
        internals.RT.RaeType:         (rae_type_implementation, rae_type_type_info),
        internals.RT.AbstractType:    (abstract_type_implementation, abstract_type_type_info),
        internals.RT.Root:            (root_imp, root_tp),
        internals.RT.Blueprint:       (blueprint_imp, None),
        internals.RT.Docstring:       (docstring_imp, None),
        internals.RT.SourceCode:      (source_code_imp, None),
        
        
        internals.RT.GetIn:           (get_in_imp, get_in_tp),
        internals.RT.InsertIn:        (insert_in_imp, insert_in_tp),
        internals.RT.Update:          (update_imp, None),
        internals.RT.UpdateIn:        (update_in_imp, update_in_tp),
        internals.RT.UpdateAt:        (update_at_imp, update_at_tp),
        internals.RT.InsertAt:        (insert_at_imp, None),
        internals.RT.RemoveIn:        (remove_in_imp, remove_in_tp),
        internals.RT.RemoveAt:        (remove_at_imp, None),
        internals.RT.Merge:           (merge_imp, merge_tp),
        internals.RT.MergeWith:       (merge_with_imp, merge_with_tp),
        internals.RT.IntToAlpha:      (int_to_alpha_imp, None),
        internals.RT.PermuteTo:       (permute_to_imp, None),
                
        internals.RT.Cycle:           (cycle_imp, cycle_tp),
        internals.RT.Repeat:          (repeat_imp, repeat_tp),
        internals.RT.Contains:        (contains_imp, contains_tp),
        internals.RT.ContainedIn:     (contained_in_imp, contained_in_tp),
        # internals.RT.Skip:            (skip_imp, skip_tp),
        internals.RT.All:             (all_imp, None),
        internals.RT.Any:             (any_imp, any_tp),
        internals.RT.Join:            (join_imp, join_tp),
        internals.RT.Trim:            (trim_imp, trim_tp),
        internals.RT.TrimLeft:        (trim_left_imp, trim_left_tp),
        internals.RT.TrimRight:       (trim_right_imp, trim_right_tp),
        internals.RT.Yo:              (yo_implementation, yo_type_info),
        
        internals.RT.Sign:            (sign_imp, sign_tp),
        internals.RT.Attempt:         (attempt_imp, attempt_tp),
        internals.RT.Bypass:          (bypass_imp, bypass_tp),
        internals.RT.Pattern:         (pattern_imp, pattern_tp),
        internals.RT.Replace:         (replace_imp, replace_tp),
        internals.RT.Distinct:        (distinct_imp, distinct_tp),
        internals.RT.DistinctBy:      (distinct_by_imp, distinct_by_tp),
        internals.RT.IsDistinct:      (is_distinct_imp, is_distinct_tp),
        internals.RT.IsDistinctBy:    (is_distinct_by_imp, is_distinct_by_tp),
        internals.RT.Shuffle:         (shuffle_imp, shuffle_tp),
        internals.RT.Slice:           (slice_imp, slice_tp),
        internals.RT.Split:           (split_imp, split_tp),
        internals.RT.SplitLeft:       (split_left_imp, None),
        internals.RT.SplitRight:      (split_right_imp, None),
        internals.RT.Graphviz:        (graphviz_imp, graphviz_tp),
        internals.RT.ToFlatGraph:     (to_flatgraph_imp, None),
        internals.RT.Parse:           (parse_imp, None),
        
        internals.RT.Always:              (always_imp, always_tp),
        
        internals.RT.WithoutAbsorbed:     (without_absorbed_imp, None),
        internals.RT.Absorbed:            (absorbed_imp, None),
        
        internals.RT.CartesianProduct:    (cartesian_product_imp, cartesian_product_tp),
        internals.RT.Permutations:        (permutations_imp, permutations_tp),
        internals.RT.Combinations:        (combinations_imp, combinations_tp),
        internals.RT.Add:                 (add_imp, add_tp),
        internals.RT.Sum:                 (sum_imp, None),
        internals.RT.Product:             (product_imp, None),
        internals.RT.Subtract:            (subtract_imp, subtract_tp),
        internals.RT.Multiply:            (multiply_imp, multiply_tp),
        internals.RT.Divide:              (divide_imp, divide_tp),
        internals.RT.Mean:                (mean_imp, mean_tp),
        internals.RT.Variance:            (variance_imp, variance_tp),
        internals.RT.Power:               (power_imp, power_tp),
        internals.RT.Exponential:         (exponential_imp, exponential_tp),
        internals.RT.Logarithm:           (logarithm_imp, logarithm_tp),
        internals.RT.Max:                 (max_imp, max_tp),
        internals.RT.Min:                 (min_imp, min_tp),
        internals.RT.Clamp:               (clamp_imp, None),
        internals.RT.MaxBy:               (max_by_imp, max_by_tp),
        internals.RT.MinBy:               (min_by_imp, min_by_tp),
        internals.RT.Equals:              (equals_imp, equals_tp),
        internals.RT.GreaterThan:          (greater_than_imp, greater_than_tp),
        internals.RT.LessThan:            (less_than_imp, less_than_tp),
        internals.RT.GreaterThanOrEqual:  (greater_than_or_equal_imp, greater_than_or_equal_to),
        internals.RT.LessThanOrEqual:     (less_than_or_equal_imp, less_than_or_equal_to),
        internals.RT.Not:                 (not_imp, not_tp),
        internals.RT.And:                 (and_imp, and_tp),
        internals.RT.Or:                  (or_imp, or_tp),
        internals.RT.Xor:                 (xor_imp, xor_tp),
        internals.RT.Peel:                (peel_imp, peel_tp),
        internals.RT.Match:               (match_imp, match_tp),
        internals.RT.MatchOn:             (match_on_imp, None),

        internals.RT.Sync:                (sync_imp, sync_tp),
        internals.RT.Tag:                 (tag_imp, tag_tp),
        internals.RT.Untag:               (untag_imp, untag_tp),        
        internals.RT.ToClipboard:         (to_clipboard_imp, to_clipboard_tp),
        internals.RT.FromClipboard:       (from_clipboard_imp, from_clipboard_tp),
        internals.RT.TextArt:             (text_art_imp, text_art_tp),

        internals.RT.ToJSON:             (to_json_imp, to_json_tp), 
        internals.RT.FromJSON:           (from_json_imp, from_json_tp),

        internals.RT.ToYaml:             (to_yaml_imp, to_yaml_tp),
        internals.RT.FromYaml:           (from_yaml_imp, from_yaml_tp),

        internals.RT.ToToml:              (to_toml_imp, to_toml_tp),
        internals.RT.FromToml:            (from_toml_imp, from_toml_tp),

        internals.RT.ToCSV:               (to_csv_imp, from_csv_tp),
        internals.RT.FromCSV:             (from_csv_imp, from_csv_tp),

        internals.RT.ReadFile:            (read_file_imp, read_file_tp),
        internals.RT.LoadFile:            (load_file_imp, load_file_tp),
        internals.RT.WriteFile:           (write_file_imp, write_file_tp),
        internals.RT.SaveFile:            (save_file_imp, save_file_tp),

        internals.RT.PandasToGd:          (pandas_to_gd_imp, pandas_to_gd_tp),
        
        internals.RT.ToPipeline:          (to_pipeline_imp, to_pipeline_tp),
        internals.RT.Inject:              (inject_imp, inject_tp),
        internals.RT.InjectList:          (inject_list_imp, inject_list_tp),

        internals.RT.ZasciiToAsg:         (zascii_to_asg_imp, zascii_to_asg_tp),
        internals.RT.ZasciiToFlatGraph:   (zascii_to_flatgraph_imp, zascii_to_flatgraph_tp,),
        internals.RT.ZasciiToBlueprintFg:  (zascii_to_blueprint_fg_imp, zascii_to_blueprint_fg_tp),

        internals.RT.ReplaceAt:           (replace_at_imp, replace_at_tp),
        internals.RT.RandomPick:          (random_pick_imp, random_pick_tp),
        internals.RT.PadLeft:             (pad_left_imp, None),
        internals.RT.PadRight:            (pad_right_imp, None),
        internals.RT.PadCenter:           (pad_center_imp, None),
        internals.RT.Floor:               (floor_imp, None),
        internals.RT.Ceil:                (ceil_imp, None),
        internals.RT.Round:               (round_imp, None),
        internals.RT.IsAlpha:             (is_alpha_imp, is_alpha_tp),
        internals.RT.IsNumeric:           (is_numeric_imp, None),
        internals.RT.IsAlphaNumeric:      (is_alpha_numeric_imp, None),
        internals.RT.IsAlpha:             (is_alpha_imp, is_alpha_tp),
        internals.RT.ToUpperCase:         (to_upper_case_imp, to_upper_case_tp),
        internals.RT.ToLowerCase:         (to_lower_case_imp, to_lower_case_tp),
        internals.RT.ToPascalCase:        (to_pascal_case_imp, to_pascal_case_tp),
        internals.RT.ToCamelCase:         (to_camel_case_imp, to_camel_case_tp),
        internals.RT.ToKebabCase:         (to_kebab_case_imp, to_kebab_case_tp),
        internals.RT.ToSnakeCase:         (to_snake_case_imp, to_snake_case_tp),
        internals.RT.ToScreamingSnakeCase:(to_screaming_snake_case_imp, to_screaming_snake_case_tp),


        internals.RT.Unpack:              (unpack_implementation, unpack_type_info),
        internals.RT.IndexesOf:           (indexes_of_imp, None),
        internals.RT.GraphSliceIndex:     (graph_slice_index_imp, None),

        internals.RT.MakeRequest:        (make_request_imp, make_request_tp),
        internals.RT.Blake3:             (blake3_imp, blake3_tp),
        internals.RT.ValueHash:          (value_hash_imp, value_hash_tp),
        internals.RT.ToZefList:          (to_zef_list_imp, to_zef_list_tp),
        internals.RT.Transact:           (transact_imp, transact_tp),

        internals.RT.Function:           (function_imp, function_tp),
        internals.RT.On:                 (on_implementation, None),
        internals.RT.Range:              (range_imp, range_tp),
        internals.RT.ZstandardCompress:  (zstandard_compress_imp, None),
        internals.RT.ZstandardDecompress:(zstandard_decompress_imp, None),        
        internals.RT.ToBytes:            (to_bytes_imp, None),
        internals.RT.Utf8bytesToString:  (utf8bytes_to_string_imp, None),
        internals.RT.Base64stringToBytes:(base64string_to_bytes_imp, None),
        internals.RT.BytesToBase64string:(bytes_to_base64string_imp, None),
        internals.RT.IsBetween:          (is_between_imp, None),
        internals.RT.If:                 (If_imp, None),
        internals.RT.Field:              (field_imp, None),
        internals.RT.Fields:             (fields_imp, None),
        internals.RT.Apply:              (apply_imp, None),
        internals.RT.SplitOnNext:        (split_on_next_imp, None),


        internals.RT.Examples:           (examples_imp, None),
        internals.RT.Signature:          (signature_imp, None),
        internals.RT.Tags:               (tags_imp, None),
        internals.RT.RelatedOps:         (related_ops_imp, None),
        internals.RT.OperatesOn:         (operates_on_imp, None), 
        internals.RT.UsedFor:            (used_for_imp, None),

        internals.RT.Gather:             (gather_imp, None),
        internals.RT.Alias:              (alias_imp, None),
        internals.RT.Splice:             (splice_imp, None),
        internals.RT.FlattenDict:        (flatten_dict_imp, None),
        internals.RT.UnflattenDict:      (unflatten_dict_imp, None),
        internals.RT.TokenName:          (token_name_imp, None),

        internals.RT.ZefID:              (zef_id_imp, None),

        internals.RT.ToObject:             (to_object_imp, None),

        internals.RT.InferTypes :          (infer_types, None),
        internals.RT.Deduplicate:          (deduplicate, None),
        internals.RT.IdentifyEntities:     (identify_entities, None),


 
        internals.RT.RecursiveFlatten:     (recursive_flatten_imp, None),
        internals.RT.SplitAt:              (split_at_imp, None),
        internals.RT.SplitLines:           (split_lines_imp, None),
        internals.RT.FilterMap:            (filter_map_imp, None),
        internals.RT.EndsWith:             (ends_with_imp, None),
        internals.RT.StartsWith:           (starts_with_imp, None),
}





# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import report_import
report_import("zef.core.graph_additions.common")

from ..VT import *
from ..VT.rae_types import RAET_get_token, RAET_get_names, RAET_without_names

from .._ops import *

from ..zef_functions import func

# from ..atom import _get_atom_id, _get_fields, _get_atom_type, _get_ref_pointer, get_all_ids, get_most_authorative_id, find_zefref
from ..atom import get_all_ids, get_most_authorative_id, find_concrete_pointer

from .types import *

# ID generation uses a counting integer to distinguish ids in the same set and a prefix (e.g. coming from a time) to try and distinguish between different runs
GenIDState = Tuple[String,Int]

##############################
# * General Utils

def get_uvt_type(uvi: UserValueInstance):
    from ..user_value_type import _user_value_type_registry
    return _user_value_type_registry[uvi._user_type_id]

# TODO: This needs to be turned into a UVT.
from dataclasses import dataclass
@dataclass(frozen=True)
class OriginallyUserID:
    obj: None

    def __repr__(self):
        return f"was:{self.obj}"

@func
def maybe_unwrap_variable(x):
    return LazyValue(x) | match[
        (Variable & Is[get_field["name"]
            | is_a[OriginallyUserID]],
         get_field["name"]
               | get_field["obj"]),
        (Any, identity)
    ] | collect

def maybe_unwrap_variables_in_receipt(receipt):
    return (receipt
               | items
               | map[apply[first | maybe_unwrap_variable,
                           second]]
               | func[dict]
               | collect)

def wrap_user_id(thing):
    if isinstance(thing, ExtraUserAllowedIDs):
        return convert_extra_allowed_id(thing)
    return SymbolicExpression(OriginallyUserID(thing))

def convert_extra_allowed_id(thing: ExtraUserAllowedIDs):
    if isinstance(thing, NamedZ):
        id = thing.root_node.arg2
    elif isinstance(thing, NamedAny):
        id = single(absorbed(thing))
    else:
        raise Exception("Shouldn't get here")
    # Tag this with OriginallyUserID so that we can unwrap it later on
    return SymbolicExpression(OriginallyUserID(id))

# def names_of_raet(raet):
#     from .types import WishID
#     return raet._d["absorbed"] | filter[WishID] | collect
# def bare_raet(raet):
#     from .types import WishID
#     # Note: need tuple as collect produces lists by default which are not what
#     # is usually in absorbed.
#     return raet._replace(absorbed=tuple(raet._d["absorbed"] | filter[Not[WishID]] | collect))
def names_of_raet(raet):
    from .types import WishID
    names = (raet
             | match[
                 (PureET | PureRT | PureAET, RAET_get_names),
                 (AtomClass, get_all_ids),
             ]
             | map[match[
                 (AllIDs, identity),
                 (Any, wrap_user_id)
                 ]]
             | collect)
    return names
def bare_raet(raet):
    if isinstance(raet, PureET | PureRT | PureAET):
        return RAET_without_names(raet)
    elif isinstance(raet, AtomClass):
        return rae_type(raet)
    else:
        raise Exception(f"Unknown type in bare_raet: {raet}")


@func
def not_implemented_error(obj, text, *others):
    raise NotImplementedError(text + f" Obj: {obj}")

@func
def wrap_list(x):
    return [x]

def generate_initial_state(label: String) -> GenIDState:
    import time
    return (label + str(time.time()), 0)

def gen_internal_id(prev_state: GenIDState) -> Tuple[WishIDInternal, GenIDState]:
    prefix,prev_id = prev_state
    next_id = prev_id + 1
    id_str = WishIDInternal(f"{prefix}_{next_id}")
    next_state = (prefix, next_id)
    return id_str, next_state

# Temporary hack for simpler prototyping
def global_gen_internal_id():
    global_gen_internal_id.last_id += 1
    return WishIDInternal(f"global_{global_gen_internal_id.last_id}")
global_gen_internal_id.last_id = 1
    

def convert_scalar(scalar):
    return AETWithValue({"aet": map_scalar_to_aet(scalar),
                         "value": Val(scalar)})

def map_scalar_to_aet(x):
    return x | match[
        (PyBool, always[AET.Bool]),
        (Int, always[AET.Int]),
        (Float, always[AET.Float]),
        (String, always[AET.String]),
        (Time, always[AET.Time]),
        (Enum, make_enum_aet),
        (QuantityFloat, make_qf_aet), 
        (QuantityInt, make_qi_aet), 
        (ValueType, always[AET.Type]),
        (UserValueInstance, lambda x: x._get_type()),
        (Any, not_implemented_error["TODO scalar type"]),
    ] | collect

def make_enum_aet(x):
    return getattr(AET.Enum, x.enum_type)

def make_qf_aet(x):
    return getattr(AET.QuantityFloat, x.unit.enum_value)

def make_qi_aet(x):
    return getattr(AET.QuantityInt, x.unit.enum_value)

def can_assign_to(val, z: AtomClass):
    if not isinstance(z, AttributeEntity):
        return False
    # Special cases
    if isinstance(val, Int):
        if isinstance(z, AET.Float):
            return True

    if isinstance(val, PrimitiveValue):
        return map_scalar_to_aet(val) == rae_type(z)

    aet = rae_type(z)
    # TODO: Need to get the type as a ValueType from aet (for simple and/or
    # complex types) and then do the isinstance check on val for that.
    from ..VT.rae_types import RAET_get_token
    if RAET_get_token(aet).complex_value is not None:
        return isinstance(val, RAET_get_token(aet).complex_value)
    else:
        raise NotImplementedError("TODO: Need to map AET to value type to compare with value")
    
def most_recent_rae_on_graph(origin_uid: EternalUID, g: Graph) -> Nil|ZefRef:
    """
    Will definitely not return a BT.ForeignInstance, always an instance.
    It could be that the node asked for has its origin on this graph 
    (the original rae may still be alive or it may be terminated)

    This function should only be called DURING a transaction, not in preparation
    for a transaction.

    Returns:
        ZefRef: this graph knows about this: found instance
        None: this graph knows nothing about this RAE
    """
    if origin_uid not in g:
        return None     # this graph never knew about a RAE with this origin uid

    zz = g[origin_uid]
    if BT(zz) in {BT.FOREIGN_ENTITY_NODE, BT.FOREIGN_ATTRIBUTE_ENTITY_NODE, BT.FOREIGN_RELATION_EDGE}:
        from ..graph_slice import get_instance_rae
        return get_instance_rae(origin_uid, now(g))
        
    elif BT(zz) in {BT.ENTITY_NODE, BT.ATTRIBUTE_ENTITY_NODE, BT.RELATION_EDGE}:
        if zz | exists_at[now(g)] | collect:
            return zz | in_frame[now(g)] | collect
        else:
            return None
    elif BT(zz) in {BT.ROOT_NODE, BT.TX_EVENT_NODE}:
        return zz | in_frame[now(g)] | collect
    else:
        raise RuntimeError("Unexpected option in most_recent_rae_on_graph")


@func
def match_rules(input: Tuple, rules: List[Tuple[ValueType,Any]]):
    # This is like match_on but also splats the input arguments out (input must
    # be a tuple and first argument is used for the match condition).

    for typ, func in rules:
        # Note: we have to use isinstance here, as LazyValues could be passed in
        # and need to be processed without evaluation.
        if isinstance(input[0], typ):
            return func(*input)
    raise Error.MatchError("No case matched", input[0])


@func
def scanmany(itr_in, reducers, maybe_init=None):
    def scanmany_inner():
        itr = iter(itr_in)
        if maybe_init is None:
            try:
                cur_item = next(itr)
            except StopIteration:
                raise Exception("scanmany: no init item")
        else:
            cur_item = maybe_init

        yield cur_item
        for new_item in itr:
            comb_item = ()
            for cur,new,func in zip(cur_item, new_item, reducers):
                comb = func(cur, new)
                comb_item = comb_item + (comb,)
            cur_item = comb_item
            yield cur_item
            
    return ZefGenerator(scanmany_inner)

@func
def reducemany(itr, reducers, maybe_init=None):
    return itr | scanmany[reducers][maybe_init] | last | collect

@func
def merge_nodups(x, y):
    shared_keys = set(keys(x)) & set(keys(y))
    assert len(shared_keys) == 0, f"Duplicates found in keys of dictionaries passed into merge_nodups: {shared_keys}"
    return merge(x,y)

def maybe_parse_uid(s) -> Nil|EternalUID:
    if isinstance(s, str) and s.startswith("㏈_"):
        if len(s) == 35:
            return internals.EternalUID.from_base64(s[:24])
        elif len(s) == 24:
            return internals.EternalUID.from_base64(s)
        else:
            raise Exception("UID string is the wrong length")

        return None
    
def force_as_id(x):
    if isinstance(x, AtomRef):
        return origin_uid(x)
    if isinstance(x, str):
        ouid = maybe_parse_uid(x)
        if ouid is not None:
            return ouid
    if isinstance(x, AllIDs):
        return x
    if isinstance(x, FlatRefUID):
        return x
    return wrap_user_id(x)


def id_preference_pair(x: AllIDs, y: AllIDs) -> AllIDs:
    # Take the dominant id from two given ids: EternalUID >> Variable >> WishIDInternal

    assert isinstance(x, AllIDs) and isinstance(y, AllIDs)
    if isinstance(x, EternalUID) or isinstance(y, EternalUID):
        if isinstance(x, EternalUID) and isinstance(y, EternalUID) and x != y:
            raise Exception("Two different EUIDs are trying to be merged!")
        if isinstance(x, EternalUID):
            return x
        return y
    elif isinstance(x, Variable) or isinstance(y, Variable):
        if isinstance(x, Variable):
            return x
        return y
    else:
        return x

def id_preference(l: List[AllIDs]) -> AllIDs:
    out = l | reduce[id_preference_pair] | collect
    return out


def find_rae_in_target(global_uid, target):
    if isinstance(target, GraphSlice):
        # TODO: This migth not be good enough, if the gs is not the same as the
        # latest slice on the graph... but then why should it not be that?
        from ..graph_slice import get_instance_rae
        return get_instance_rae(global_uid, target)
    elif isinstance(target, FlatGraph):
        if global_uid in target:
            return target[global_uid]
        else:
            return None
    else:
        raise Exception(f"Don't understand target: {target}")


def find_value_node_in_target(val, glike):
    # This requires a bit of distinguishing between graphs and flat graphs,
    # because graphs need to serialize first.

    if isinstance(glike, GraphSlice):
        s = internals.val_as_serialized_if_necessary(val)
        return Graph(glike).get_value_node(s)
    elif isinstance(glike, FlatGraph):
        if val in glike:
            return glike[val]
        return None
        


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .VT import *
from .VT import insert_VT, make_VT

def Atom_is_global_identifier(input):
    if type(input) == str and input.startswith("㏈-"):
        return True
    if isinstance(input, EternalUID):
        return True
    # TODO: This will go away at some point
    if isinstance(input, ZefRefUID):
        return True
    if isinstance(input, DBStateRefUID):
        return True
    if isinstance(input, FlatRefUID):
        return True
    if isinstance(input, DelegateRef):
        return True
    return False


# This is like a GraphSlice but without an explicit ZefRef tx.
DBStateUID = UserValueType("DBStateUID",
                           Dict,
                           Pattern[{"tx_uid": BaseUID,
                                    "graph_uid": BaseUID}],
                           forced_uid="5f58a60ab5ae121b")

def make_dbstate_uid(gs):
    from ._ops import uid
    ctx = gs.tx
    tx_uid = uid(ctx).blob_uid
    graph_uid = uid(ctx).graph_uid
    return DBStateUID(tx_uid=tx_uid, graph_uid=graph_uid)

def find_dbstate(dbstate_uid: DBStateUID):
    # Tries to find the DBState for the given UID but returns None if we can't
    # without causing a graph load.
    gref = GraphRef(dbstate_uid.graph_uid)
    from .internals import get_loaded_graph
    g = get_loaded_graph(gref)
    if g is None:
        return None
    gs = GraphSlice(g[dbstate_uid.tx_uid])
    return gs

    
# A temporary internal structure to represent a global id with a reference frame
DBStateRefUID = UserValueType("DBStateRefUID",
                           Dict,
                           Pattern[{"global_uid": EternalUID | DelegateRef,
                                    "dbstate_uid": DBStateUID}],
                           forced_uid="e2481c246411fd50")

def make_dbstateref_uid(z):
    if isinstance(z, Atom_):
        z = get_ref_pointer(z)
    if not isinstance(z, ZefRef):
        raise Exception(f"Can't get DBStateRefUID from a {z}")

    from ._ops import origin_uid, frame, get_field, collect, uid

    global_uid = origin_uid(z)
    # Have to avoid calling to_tx as that would recurse into this function.
    dbstate_uid = make_dbstate_uid(z | frame | collect)

    return DBStateRefUID(
        global_uid=global_uid,
        dbstate_uid=dbstate_uid,
    )

FlatGraphPlaceholder = UserValueType("FlatGraphPlaceholder", Int, Any)
FlatRefUID = UserValueType("FlatRefUID",
                           Dict,
                           Pattern[{"idx": Int,
                                    "flatgraph": FlatGraphPlaceholder}])

def make_flatref_uid(fr):
    from .flat_graph import register_flatgraph
    h = register_flatgraph(fr.fg)
    placeholder = FlatGraphPlaceholder(h)
    return FlatRefUID(idx=fr.idx, flatgraph=placeholder)

# UIDString = String & Where[startswith["㏈-"]]
UIDString = String & Is[lambda x: x.startswith["㏈-"]]
AtomIdentity = (
    Pattern[{
        Optional["global_uid"]: EternalUID,
        Optional["frame_uid"]: DBStateUID,
        Optional["local_names"]: List[~UIDString],
        Optional["flatref_idx"]: Int,
    }]
    # & Cond[Contains["frame_uid"]][Contains["global_uid"]]
    & Cond[Is[lambda x: "frame_uid" in x]]
          [Is[lambda x: "global_uid" in x] | Is[lambda x: "flatref_idx" in x]]
)

# Yes, the last bit is just List[Any] but it makes more sense for me to write
# down the various parts.
PrettyAtomIdentity = AtomIdentity | List[UIDString | EternalUID | ~UIDString]



    

def interpret_atom_identity(inputs) -> AtomIdentity:
    # Danny has been writing this from a combination of memory and making up
    # things where needed. Here is a list of points that need addressing.
    # TODO:
    # - what should "global identifier" be called?
    # - EternalUID should be replaced with new global identifier binary object
    # - tx_uid and graph_uid should be replaced with their appropriate sizes.

    from ._ops import uid, filter, collect, single
    from .flat_graph import FlatGraph, FlatRef, FlatRef_maybe_uid

    # Given a tuple of inputs to an Atom, determine a standardised dictionary to represent this.

    # Only one item in the list is allowed to be an item with a global
    # identifier (for now?). So check this first.
    glob_identifiers = []
    others = set()
    for input in inputs:
        if Atom_is_global_identifier(input):
            glob_identifiers += [input]
        else:
            others.add(input)
    if len(glob_identifiers) >= 2:
        raise Exception("Can't interpret atom identity, too many global identifiers")
        
    desc = {}
    
    if len(glob_identifiers) == 1:
        input = glob_identifiers[0]
        if type(input) == str and input.startswith("㏈-"):
            parts = input[len("㏈-"):].split("-")

            def parse_global_uid(x):
                out = uid(x)
                assert isinstance(out, EternalUID)
                return out
            
            # Option 1: just a global identifier "㏈-49836587346876342856236478"
            if len(parts) == 1:
                # TODO: this type of ID will change
                desc = dict(global_uid=parse_global_uid(parts[0]))
            # Option 2: "㏈-49836587346876342856236478-xxx-yyy"
            elif len(parts) == 3:
                dbstate_uid = DBStateUID(
                    tx_uid=BaseUID(parts[1]),
                    graph_uid=BaseUID(parts[2])
                )
                desc = dict(
                    global_uid=parse_global_uid(parts[0]),
                    frame_uid=dbstate_uid,
                )
            # Option 3: "㏈-49836587346876342856236478-FG#xxx"
            elif len(parts) == 2:
                assert parts[1].startswith("FG#")
                flatgraph_uid = FlatGraphPlaceholder(int(parts[1][len("FG#"):]))
                desc = dict(
                    global_uid=parse_global_uid(parts[0]),
                    frame_uid=flatgraph_uid,
                )
            else:
                raise Exception(f"Not sure how to interpret this kind of identifier! {input}")
        elif isinstance(input, EternalUID):
            desc = dict(global_uid=input)
        elif isinstance(input, ZefRefUID):
            raise Exception("Shouldn't be getting a ZefRefUID anymore - we would need to load the graph to determine the global uid")
        elif isinstance(input, DBStateRefUID):
            desc = dict(
                global_uid=input.global_uid,
                frame_uid=input.dbstate_uid,
            )
        elif isinstance(input, FlatRefUID):
            # The global identifier must be given in the FlatRef if we hit this point
            desc = dict(
                flatref_idx=input.idx,
                frame_uid=input.flatgraph,
            )
        elif isinstance(input, DelegateRef):
            desc = dict(global_uid=input)
        else:
            raise Exception("Shouldn't get here")
    elif len(glob_identifiers) >= 2:
        raise Exception("Can't have more than one global identifier")
    else:
        pass

    # There could be a FlatGraph in the list when strictly_invertible is used in
    # pretty_atom_identity. This is to be interpreted as the reference frame
    fgs = others | filter[FlatGraphPlaceholder] | collect
    others = others | filter[~FlatGraphPlaceholder] | collect
    if len(fgs) > 0:
        if "frame_uid" in desc:
            raise Exception("Can't have a FlatGraph as a reference frame when the atom already has a reference graph")
        desc["frame_uid"] = single(fgs)


    # Same but with DBStateUID
    dbs = others | filter[DBStateUID] | collect
    others = others | filter[~DBStateUID] | collect
    if len(dbs) > 0:
        if "frame_uid" in desc:
            raise Exception("Can't have a DBState as a reference frame when the atom already has a reference graph")
        desc["frame_uid"] = single(dbs)


    if len(others) > 0:
        desc["local_names"] = others

    return desc

def pretty_atom_identity(desc: AtomIdentity):
    from .flat_graph import FlatGraph

    assert isinstance(desc, AtomIdentity)
    # The inverse of interpret_atom_identity
    names = []
    if "frame_uid" not in desc:
        if "global_uid" in desc:
            if isinstance(desc["global_uid"], DelegateRef):
                # We can't pretty print delegates so just output directly
                names += [desc["global_uid"]]
            else:
                names += ["㏈-" + str(desc["global_uid"])]

    elif isinstance(desc["frame_uid"], DBStateUID):
        if isinstance(desc["global_uid"], DelegateRef):
            # We can't pretty print delegates so just output directly
            names += [DBStateRefUID(global_uid=desc["global_uid"],
                                    dbstate_uid=desc["frame_uid"])]
        else:
            s = "㏈-" + str(desc["global_uid"])
            s += f"-{desc['frame_uid'].tx_uid}-{desc['frame_uid'].graph_uid}"
            names += [s]

    elif isinstance(desc["frame_uid"], FlatGraphPlaceholder):
        if "global_uid" in desc:
            s = "㏈-" + str(desc["global_uid"])
            s += f"-FG#{desc['frame_uid']._value}"
            names += [s]
        else:
            names += [FlatRefUID(global_uid=desc["flatref_idx"],
                                frame_uid=desc["frame_uid"])]
            
    else:
        raise NotImplementedError(f"Unknown type of frame: {desc['frame_uid']}")

    if "local_names" in desc:
        names += list(desc["local_names"])

    return tuple(names)
    

class Atom_:
    """
    AtomType: e.g. ET.Person, RT.Foo, AET[String]
    names: containing elements like e.g. '㏈-49836587346-57863478568734-876342856236478' / 'Fred'
    fields: Dict
    # not for now: optional_zefref_or_ezefref

    derived quantity reference_type: e.g. 'unidentified', 'local_name', 'red', 'db_ref', 'db_state_ref', 'graph_ref'
    """

    def __init__(self, *args, **fields):
        # Construct a blank slate and then call ourselves
        object.__setattr__(self, "atom_type", None)
        object.__setattr__(self, "atom_id", {})
        object.__setattr__(self, "fields", {})
        object.__setattr__(self, "ref_pointer", None)

        # Special case for an empty atom
        if len(args) == 0 and len(fields) == 0:
            return
        
        # Because we can't return a completely different value, we have to
        # assign everything to us.
        new_atom = self(*args, **fields)

        attrs = ["atom_type", "atom_id", "fields", "ref_pointer"]
        for attr in attrs:
            object.__setattr__(self, attr, object.__getattribute__(new_atom, attr))
    

    def __replace__(self, **kwargs):
        attrs = ["atom_type", "atom_id", "fields", "ref_pointer"]
        assert all(kwarg in attrs for kwarg in kwargs), "Trying to set an Attribute for Atom that isn't allowed."

        new_atom = Atom_()
        for attr in attrs:
            if attr in kwargs:
                object.__setattr__(new_atom, attr, kwargs[attr])
            else:
                object.__setattr__(new_atom, attr, object.__getattribute__(self, attr))

        return new_atom

    def __call__(self, *args, **kwargs):
        atom_type = object.__getattribute__(self, "atom_type")
        fields = dict(object.__getattribute__(self, "fields"))
        ref_pointer = object.__getattribute__(self, "ref_pointer")

        # We represent our internal id as a pretty identity so that we can
        # conform to whatever the user is speaking.
        atom_id = object.__getattribute__(self, "atom_id")
        names = pretty_atom_identity(atom_id)

        fields.update(kwargs)
        
        from ._ops import is_a, rae_type, source, target, origin_uid, abstract_type, discard_frame, value, uid
        from .VT import BlobPtr, RAET, EntityRef, RelationRef, AttributeEntityRef, FlatRef, Relation, AttributeEntity
        from .flat_graph import FlatRef, FlatRef_rae_type, FlatRef_maybe_uid

        for arg in args:
            if is_a(arg, AtomClass):
                # Take over everything about the other atom
                other_ref_pointer = object.__getattribute__(arg, "ref_pointer")
                other_fields = object.__getattribute__(arg, "fields")
                other_atom_type = object.__getattribute__(arg, "atom_type")
                other_atom_id = object.__getattribute__(arg, "atom_id")

                assert ref_pointer is None or other_ref_pointer is None
                if ref_pointer is None:
                    ref_pointer = other_ref_pointer

                fields.update(other_fields)

                assert atom_type is None or other_atom_type is None
                if atom_type is None:
                    atom_type = other_atom_type

                other_names = pretty_atom_identity(other_atom_id)
                names = names + other_names
            
            elif is_a(arg, BlobPtr):
                # This means we can extract the atom_type and uid from the Ref
                assert ref_pointer is None
                ref_pointer = arg
                ptr_atom_type = abstract_type(ref_pointer)
                assert atom_type is None or atom_type == ptr_atom_type
                atom_type = ptr_atom_type
                if isinstance(arg, ZefRef):
                    names = (make_dbstateref_uid(ref_pointer), *names)
                else:
                    names = (origin_uid(ref_pointer), *names)
        
            elif is_a(arg, FlatRef):
                ref_pointer = arg

                from .flat_graph import register_flatgraph
                fr_uid = make_flatref_uid(arg)
                tag = FlatRef_maybe_uid(arg)
                if isinstance(tag, EternalUID):
                    names = (tag, fr_uid.flatgraph, *names)
                else:
                    names = (fr_uid, *names)
                atom_type = FlatRef_rae_type(arg)

            elif is_a(arg, EntityRef | AttributeEntityRef | RelationRef):
                rae = arg
                assert atom_type is None
                atom_type = rae_type(rae)
                names = (origin_uid(rae), *names)

            elif is_a(arg, RAET):
                assert atom_type is None
                atom_type = arg
            else:
                names = names + (arg,)
        
        atom_id = interpret_atom_identity(names)

        return self.__replace__(atom_type=atom_type,
                                atom_id=atom_id,
                                fields=fields,
                                ref_pointer=ref_pointer)


    def __repr__(self):
        ref_pointer = _get_ref_pointer(self)
        atom_type = _get_atom_type(self)
        atom_id = pretty_atom_identity(_get_atom_id(self))
        fields = _get_fields(self)
        # items = [f'"{get_reference_type(self)}"'] + [f"{k}={v!r}" for k,v in fields.items()]
        items = [f"{k}={v!r}" for k,v in fields.items()]
        items = [repr(name) for name in atom_id] + list(items)
        if ref_pointer:
            # This is just for us while we are developing. Will be removed in the future.
            items += ["**{}"]
        return f'{atom_type}({f", ".join(items)})'

    def __setattr__(self, name, value):
        raise AttributeError("Atoms are immutable")
    def __delattr__(self, name):
        raise AttributeError("Atoms are immutable")

    def __getattribute__(self, name):
        if name.startswith("__"):
            return object.__getattribute__(self, name)
        # Need to convert KeyErrors to AttributeErrors for calling code to handle things like __x__ accesses
        try:
            return object.__getattribute__(self, "fields")[name]
        except KeyError:
            raise AttributeError(name)
    def __dir__(self):
        return dir(object.__getattribute__(self, "fields"))

    def __eq__(self, other):
        if type(other) != Atom_:
            return False
        return (_get_atom_type(self) == _get_atom_type(other)
                and _get_ref_pointer(self) == _get_ref_pointer(other)
                and _get_atom_id(self) == _get_atom_id(other)
                and _get_fields(self) == _get_fields(other))

    def __hash__(self):
        from .VT.value_type import hash_frozen
        return hash_frozen(("Atom_", _get_atom_type(self), _get_ref_pointer(self), _get_atom_id(self), _get_fields(self)))


# This means we are really specifically just the atom class, not a generic
# "atom".
AtomClass = make_VT('AtomClass', pytype=Atom_)

def _get_atom_type(atom: AtomClass):
    return object.__getattribute__(atom, "atom_type")
def _get_atom_id(atom: AtomClass):
    return object.__getattribute__(atom, "atom_id")
def _get_fields(atom: AtomClass):
    return object.__getattribute__(atom, "fields")
def _get_ref_pointer(atom: AtomClass):
    return object.__getattribute__(atom, "ref_pointer")

def get_most_authorative_id(atom: AtomClass):
    atom_id = _get_atom_id(atom)
    if "global_uid" in atom_id:
        return atom_id["global_uid"]
    if "flatref_idx" in atom_id:
        return FlatRefUID(idx=atom_id["flatref_idx"], flatgraph=atom_id["frame_uid"])
    if "local_names" in atom_id:
        from ._ops import sort, first, collect
        from .VT.value_type import hash_frozen
        return atom_id["local_names"] | sort[hash_frozen] | first | collect
    return None

def get_all_ids(atom: AtomClass):
    atom_id = _get_atom_id(atom)

    names = []
    if "global_uid" in atom_id:
        names += [atom_id["global_uid"]]
    if "local_names" in atom_id:
        names += atom_id["local_names"]

    return tuple(names)

def find_concrete_pointer(atom: AtomClass):
    # Find the ZefRef for an atom if we can. If we can't, return None
    z = _get_ref_pointer(atom)
    if z is not None:
        if isinstance(z, EZefRef):
            return None
        return z

    atom_id = _get_atom_id(atom)
    if "frame_uid" not in atom_id:
        return None
    frame_uid = atom_id.frame_uid
    if isinstance(frame_uid, DBStateUID):
        gs = find_dbstate(frame_uid)
        if gs is None:
            return None
        z = gs[atom_id.global_uid]
        return z
    elif isinstance(frame_uid, FlatGraphUID):
        fg = lookup_flatgraph(frame_uid.flatgraph)
        fr = fg[frame_uid.tag]
        return fr
    else:
        return None


def get_uid_type(uid_str: str) -> str:
    uid_chunk_size = 12
    uid_str = uid_str.strip("㏈-")

    if len(uid_str) == (uid_chunk_size*3 + 2)   and uid_str.count("-") == 2:      # 12 * 3 + (-,-) i.e ㏈-2ae69e6f28ab-2ae69e6f28ab-2ae69e6f28ab
        return 'db_state_ref'
    elif len(uid_str) == (uid_chunk_size*2 + 1) and uid_str.count("-") == 1:    # 12 * 2 + (-) i.e ㏈-2ae69e6f28ab-2ae69e6f28ab
        return 'db_ref'
    elif len(uid_str) == (uid_chunk_size*2):        # 12 * 2 i.e ㏈-2ae69e6f28ab2ae69e6f28ab
        return 'ref'
    else:
        raise ValueError(f"Invalid UID: {uid_str}")

def uid_str_to_uid(uid_str: str) -> UID:
    uid_str = uid_str[2:]
    if get_uid_type(uid_str) == "db_state_ref":
        return ZefRefUID(*uid_str.split('-'))
    elif get_uid_type(uid_str) == "db_ref":
        return EternalUID(*uid_str.split('-'))
    elif get_uid_type(uid_str) == "ref":
        return BaseUID(uid_str)
    else:
        raise Exception(f"Unknown type of uid: {uid_str}")

def get_reference_type(atom: AtomClass) -> str:
    atom_id = get_atom_id(atom)
    if "local_names" not in atom_id and "global_uid" not in atom_id:
        return 'unidentified'
    if "global_uid" not in atom_id:
        return "local_name"
    # return get_uid_type(uids[0])
    if "tx_uid" in atom_id:
        return "db_state_ref"
    else:
        return "db_ref"


def RelationAtom_is_a(x, typ):
    from ._ops import rae_type, is_a
    return isinstance(x, AtomClass & Is[rae_type | is_a[RT]])
RelationAtom = make_VT("RelationAtom", is_a_func=RelationAtom_is_a)
def EntityAtom_is_a(x, typ):
    from ._ops import rae_type, is_a
    return isinstance(x, AtomClass & Is[rae_type | is_a[ET]])
EntityAtom = make_VT("EntityAtom", is_a_func=EntityAtom_is_a)
def AttributeEntityAtom_is_a(x, typ):
    from ._ops import rae_type, is_a
    return isinstance(x, AtomClass & Is[rae_type | is_a[AET]])
AttributeEntityAtom = make_VT("AttributeEntityAtom", is_a_func=AttributeEntityAtom_is_a)
def TXNodeAtom_is_a(x, typ):
    from ._ops import abstract_type, equals
    return isinstance(x, AtomClass & Is[abstract_type | equals[BT.TX_EVENT_NODE]])
TXNodeAtom = make_VT("TXNodeAtom", is_a_func=TXNodeAtom_is_a)
def RootAtom_is_a(x, typ):
    from ._ops import abstract_type, equals
    return isinstance(x, AtomClass & Is[abstract_type | equals[BT.ROOT_NODE]])
RootAtom = make_VT("RootAtom", is_a_func=RootAtom_is_a)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import report_import
report_import("zef.core.colors")
from .VT import *


def valid_hex_color_code(color):
    if len(color) not in {6,7}:
        return False
    try:
        if len(color) == 7: color = color[1:]
        int(color, 16)
    except ValueError:
        return False
    return True
    
ValidHEXColorCode = Is[valid_hex_color_code]
Color = UserValueType("Color", String, ValidHEXColorCode, forced_uid="10252a53a03086b7")


colors = {
    "red": Color("#E45D51"),
    "pink": Color("#FF647C"),
    "magenta": Color("#E13383"),
    "purple": Color("#AE2564"),
    "violet": Color("#BE52F3"),
    "lavender": Color("#DAA5F5"),
    "blue": Color("#A6AFFC"),
    "turquoise": Color("#4AA9C5"),
    "cyan": Color("#449ED4"),
    "indigo": Color("#3E66FB"),
    "teal": Color("#3C8A8A"),
    "green": Color("#01C48C"),
    "mint": Color("#7DE0C4"),
    "lime": Color("#3CC13B"),
    "yellow": Color("#FECF5B"),
    "gold": Color("#E3B94A"),
    "peach": Color("#FFE29E"),
    "orange": Color("#FFA26C"),
    "coral": Color("#FFC7A6"),
    "salmon": Color("#FDAFBC"),
}




# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .._ops import *
from ..VT import *

def graph_take_transactor_role_handler(eff: dict):
    from ...pyzef.main import make_primary
    make_primary(eff["graph"], True)
    return {}

def graph_release_transactor_role_handler(eff: dict):
    from ...pyzef.main import make_primary
    make_primary(eff["graph"], False)
    return {}

def graph_transaction_handler(eff: dict):
    """[summary]

    Args:
        eff (dict): {
            "type": FX.Graph.Transact,
            "target_glike": g1: GraphRef|FlatGraph,
            "commands": list_of_commands
        }
    """


    if isinstance(eff["target_glike"], FlatGraph) or (isinstance(eff["target_glike"], GraphRef) and internals.is_transactor(eff["target_glike"])):
        # For backwards compatibility we dispatch on whether this is a new or old transact effect

        if "level2_commands" in eff:
            # This is the new path
            from ..graph_additions.wish_translation2 import generate_level1_commands
            if eff["translation_rules"] == "default":
                pass
            else:
                raise Exception("There are no options for translation rules at the moment. TODO: Need to replace with distinguish/recombine/relabel/cull rules later.")

            target_ref = eff["target_glike"]
            if isinstance(target_ref, GraphRef):
                target_ref = now(Graph(target_ref))
            lvl1_cmds = generate_level1_commands(eff["level2_commands"]["cmds"], target_ref)

            if isinstance(target_ref, GraphSlice):
                from ..graph_additions.low_level import perform_level1_commands
                glike_frame,receipt = perform_level1_commands(lvl1_cmds, eff.get("keep_internal_ids", False))
            else:
                assert isinstance(target_ref, FlatGraph)
                from ..op_implementations.flatgraph_implementations import fg_transaction_implementation
                glike_frame,receipt = fg_transaction_implementation(lvl1_cmds["cmds"], target_ref)

            # TODO: Postprocess using custom info stored in the level2 commands info
            post_transact_rule = eff.get("post_transact_rule", None)
            if post_transact_rule is not None:
                receipt = post_transact_rule(receipt, lvl2_custom=eff["level2_commands"].get("custom", None))
        else:
            from ..graph_delta import perform_transaction_commands
            receipt = perform_transaction_commands(eff['commands'], eff['target_glike'])
    else:
        from ..overrides import merge
        # TODO: This will have to be updated to handle sending GraphSlices over the wire
        glike_frame,receipt = merge(eff["commands"], eff["target_glike"])
    
    # Handling receipt response needs a similar dispatch between old and new
    if "level2_commands" in eff:
        # New graph wish path

        # TODO:
        #if isinstance(glike_frame, DBStateRef):
        #    glike_frame = ... load GraphSlice from ref

        from ..graph_additions.shorthand import unpack_receipt
        if 'unpacking_template' in eff:
            return unpack_receipt(eff['unpacking_template'], receipt, glike_frame)
        elif glike_frame is not None:
            # Since g is loaded, we can return Atoms instead of AtomRefs and FlatRefs instead of FlatRefUIDs.
            if isinstance(glike_frame, GraphSlice):
                from ..atom import _get_ref_pointer
                conversions = [
                    (AtomRef,
                     lambda x: glike_frame | get[x] | collect),
                    (Atom & Is[lambda x: _get_ref_pointer(x) is None],
                     lambda x: glike_frame | get[x] | collect),
                ]
            else:
                from ..graph_additions.types import FlatRefUID
                conversions = [(FlatRefUID, lambda x: glike_frame[x.idx])]
            from .. import func
            receipt = (receipt
                        | items
                        | map[apply[first,
                                    second | match[[
                                        *conversions,
                                        (Any, identity)
                                    ]]]]
                        | func[dict]
                        | collect)
        return glike_frame, receipt
    else:
        from ..graph_delta import perform_transaction_commands, filter_temporary_ids, unpack_receipt
        # we need to forward this if it is in the effect
        if 'unpacking_template' in eff:
            return unpack_receipt(eff['unpacking_template'], receipt)
        receipt = filter_temporary_ids(receipt)
        return receipt
    

def graph_load_handler(eff: dict):
    # TODO: Make the underlying function be asynchronous and then update this to return an awaitable

    from ...pyzef.main import load_graph
    kwargs = {}
    if "mem_style" in eff:
        kwargs["mem_style"] = eff["mem_style"]
    if "progress_stream" in eff:
        kwargs["callback"] = lambda progress,stream=eff["progress_stream"]: progress | push[stream] | run
    if "progress_callback" in eff:
        kwargs["callback"] = eff["progress_callback"]

    g = load_graph(eff["tag_or_uid"], **kwargs)

    return {"g": g}


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import report_import
report_import("zef.core.graph_additions.wish_translation")

from .types import *
from .common import *



##########################################
# * Level 1 generation
#----------------------------------------

class CouldNotProcess: pass

WishMapping = Dict[WishID][EZefRef|WishID]
CmdDispatchContext = UserValueType("Level1CmdDispatchContext",
                                   Dict,
                                   Pattern[{"mapping": WishMapping, "gen_id_state": GenIDState}])

DispatchOutput = SetOf(CouldNotProcess) | Tuple[List[PleaseCommandLevel1], List[PleaseCommandLevel2], CmdDispatchContext]

# Rules should follow the signature below, but we can't specify that just yet.
# TranslationFunc = Func[PleaseCommandLevel2, CmdDispatchContext][DispatchOutput]
TranslationFunc = Any

# Really want to use Subtype[PleaseCommandLevel2] instead of ValueType
def generate_level1_commands(commands: List[PleaseCommandLevel2], gs: GraphSlice, rules: List[Tuple[ValueType,TranslationFunc]]) -> Level1CommandInfo:
    # For each command:
    #
    # - if PleaseInstantiate
    #   - see if already exists (given origin_uid)

    # - At end, consolidate such that commands which do the same thing, but might have a different WishID, are combined.

    # Note: we could save the level2 command that generated each level1 command
    # for tracability of errors back to their source.

    # - output_cmds is the order of the cmds to put into the output - these
    #   entries are unnamed
    # - cmd_naming is an efficient way to "translate" the unnamed commands into
    #   the same command with all ids applied.
    cmd_naming = {}
    output_cmds = []

    # If any commands are elided or otherwise referencing existing objects, we
    # can indicate them here. These replacements only occur for targets of
    # commands, not the identity of a command itself.
    #
    # If the mapping maps to a WishID, this indicates the target will be created
    # during the graph transaction. It guarantees that the item does not yet
    # exist on the graph.
    context = CmdDispatchContext({
        "mapping": {},
        "gen_id_state": generate_initial_state("lvl1"),
    })
    
    # We have to have multiple passes at the commands as each command can slowly
    # describe different wish ids.

    todo = list(commands)
    while len(todo) > 0:
        new_todo = []
        processed_one = False

        for cmd in todo:
            # new_lvl1_cmds,new_lvl2_cmds,mapping = cmd | match[
            result = (cmd, gs, context) | match_rules[[
                *rules,
                (Any, lambda x: not_implemented_error["Unknown cmd type"]), 
            ]] | collect

            if result == CouldNotProcess:
                new_todo.append(cmd)
            else:
                processed_one = True
                new_lvl1_cmds,new_lvl2_cmds,context = result
                for new_cmd in new_lvl1_cmds:
                    cmd_naming,output_cmds = add_command(new_cmd, cmd_naming, output_cmds)
                new_todo += new_lvl2_cmds

        if not processed_one:
            raise Exception("We were unable to process all lvl2 commands into lvl1 commands.")

        todo = new_todo

    # TODO: Go through and replace wish ids with euids where possible.
    # TODO: Then update cmd_naming so that these are consolidated into one.

    named_output_cmds = output_cmds | map[match[
        (PleaseCommandLevel1Distinguishable, lambda x: cmd_naming[x]),
        (Any, identity)
    ]] | collect

    return Level1CommandInfo({"gs": gs,
                                "cmds": named_output_cmds})
        

################################################################
# ** Dispatched instantiation cmds

# Note: these dispatch methods are meant to be immutable, but will actually
# mutate the state (context) for speed at the moment.

@func
def cmds_for_instantiate(cmd: PleaseInstantiate, gs: GraphSlice, context: CmdDispatchContext) -> DispatchOutput:
    # First look to see if we can discard this command because the item is
    # already on the graph.
    maybe_ezr = None
    if isinstance(cmd, PleaseInstantiateDistinguishable):
        if "origin_uid" in cmd:
            if cmd.origin_uid in gs:
                maybe_ezr = gs[cmd.origin_uid]
        elif isinstance(cmd.atom, PleaseInstantiateValueNode):
            maybe_ezr = find_value_node_in_target(cmd.atom, gs)
        elif isinstance(cmd.atom, PleaseInstantiateDelegate):
            maybe_ezr = delegate_of(cmd.atom, g, False)
        else:
            raise Exception("Shouldn't get here")

    if maybe_ezr is not None:
        # We have an existing item, don't need to action this but do store the
        # found item for looking up wish ids.
        if "internal_ids" in cmd:
            for wish_id in cmd.internal_ids:
                # TODO: immutable version
                context.mapping[wish_id] = maybe_ezr
        # We can discard this command.
        return [], [], context

    # If this was not already on the graph, then leave an indication that this
    # will be generated for anyone reference our wish id.
    if "internal_ids" in cmd:
        for wish_id in cmd.internal_ids:
            context.mapping[wish_id] = wish_id

    if isinstance(cmd.atom, PleaseInstantiateRelation):
        # We attempt to apply this straight away, but we can only do that if the
        # source/target is in the mapping. If they are direct atom refs, we must
        # emit merge cmds for those too.
        lvl2_cmds = []
        lvl1_cmds = []
        rt = cmd.atom["rt"]
        can_apply_now = True
        if isinstance(cmd.atom["source"], AtomRef):
            can_apply_now = False
            source_id,context.gen_id_state = gen_internal_id(context.gen_id_state)
            lvl2_cmds.append(please_instantiate_from_atom_ref(cmd.atom["source"], source_id))
        else:
            if cmd.atom["source"] not in context.mapping:
                can_apply_now = False
            source_id = cmd.atom["source"]
        if isinstance(cmd.atom["target"], AtomRef):
            can_apply_now = False
            target_id,context.gen_id_state = gen_internal_id(context.gen_id_state)
            lvl2_cmds.append(please_instantiate_from_atom_ref(cmd.atom["target"], target_id))
        else:
            if cmd.atom["target"] not in context.mapping:
                can_apply_now = False
            target_id = cmd.atom["target"]

        new_cmd = PleaseInstantiate({"atom": {"rt": rt,
                                              "source": source_id,
                                              "target": target_id}})
        if can_apply_now:
            lvl1_cmds.append(new_cmd)
        else:
            lvl2_cmds.append(new_cmd)

        return lvl1_cmds, lvl2_cmds, context
    else:
        return [cmd], [], context

@func
def cmds_for_terminate(cmd: PleaseTerminate, gs: GraphSlice, context: CmdDispatchContext) -> DispatchOutput:
    if isinstance(cmd.target, WishID):
        if cmd.target not in context["mapping"]:
            # We return this command to the pool of lvl 2 commands to translate.
            return CouldNotProcess
        target = context["mapping"][cmd.target]
    else:
        if cmd.target not in gs:
            target = None
        else:
            target = cmd.target

    # Look to see if we can discard this command because the item is
    # not in the graph slice.
    if target is None:
        # Note: if internal_ids is introduce to PleaseTerminate, then this logic will have to be updated.
        return [], [], context
    return [cmd], [], context

@func
def cmds_for_assign(cmd: PleaseAssign, gs: GraphSlice, context: CmdDispatchContext) -> DispatchOutput:
    if isinstance(cmd.target, WishID):
        if cmd.target not in context["mapping"]:
            # We return this command to the pool of lvl 2 commands to translate.
            return CouldNotProcess
        target = context["mapping"][cmd.target]
        cmd = cmd._get_type()(cmd._value | insert["target"][target] | collect)
    else:
        if cmd.target not in gs:
            raise Exception(f"PleaseAssign for a target ({cmd.target}) which is not in the graph.")

    return [cmd], [], context


default_translation_rules = [
    (PleaseInstantiate, cmds_for_instantiate),
    (PleaseTerminate, cmds_for_terminate),
    (PleaseAssign, cmds_for_assign),
]


##############################
# ** Level1 utils 

def is_command_distinguishable(cmd: PleaseCommandLevel1):
    # Distinguishable commands are those with uids attached, e.g. a ZefRef merged in.
    # Indistinguishable commands are others like ET.Machine, which will be assign different uids later.
    return 

def get_unnamed_command(cmd: PleaseCommandLevel1):
    # All but internal_id must be the same
    wo_id = cmd._value | without[["internal_ids"]] | collect
    return get_uvt_type(cmd)(wo_id)

def get_command_ids(cmd: PleaseCommandLevel1):
    return cmd._value | get["internal_ids"][[]] | collect

def add_ids(cmd: PleaseCommandLevel1, extra_names: List[WishID]):
    if len(extra_names) == 0:
        return cmd
    internal_ids = get_command_ids(cmd) + extra_names
    d = cmd._value | insert["internal_ids"][internal_ids] | collect
    return get_uvt_type(cmd)(d)

# We abstract away the adding of commands so we can consolidate the wish ids
# at the same time. This is in contrast to doing it all at once at the end
# using a group_by or equivalent, with the slight advantage that ordering is
# preserved more easily this way.
# def add_command(cmd: PleaseCommandLevel1, cmd_naming, output_cmds):
#     # TODO: THIS FUNCTION NEEDS TO RECONSIDER WHAT MIGHT BE THE SAME BY USING COLLAPSE VIA NAMES
#     #
#     # Basically we need a bipartite-graph-like structure to resolve unique
#     # objects from various names. As new commands are entered, a
#     # (internally-ided) list of actions is created and a mapping of names to
#     # that list is kept. As either a) multiple names are used to refer to two
#     # different objects via name aliasing, or b) a single name is used to refer
#     # to multiple objects, then these objects need to be reconciled into one
#     # object.
#     #
#     # From another point of view, we deliberate make all objects
#     # distinguishable, in order to track them with user-provided names. A
#     # concrete example would be that Entity(uid="abcd") is already
#     # distinguishable, but ET.Machine would internally be given an id
#     # ET.Machine[InternalID("x1")] which makes it distinguishable.
#     #
#     # However, the recombination rules are different. Two internal ids can be
#     # merged together if it turns out they are meant to be the same. An "allowed
#     # collapse". But two EUIDs cannot be collapsed together. Maybe this should
#     # be called an "external constraint". But it is really a policy we are
#     # choosing, to allow cases like: ET.X[1], ET.X[2], ET.X[1][2] instead of
#     # throwing an error.

#     # Being careful to allow for a immutable+optimise switch here. But
#     # mutating for speed for now.
#     if not isinstance(cmd, PleaseCommandLevel1Distinguishable):
#         # Indistinguishable commands should just be entered as is.
#         output_cmds.append(cmd)
#         return cmd_naming, output_cmds
#     # Distinguishable commands should be combined.
#     cmd_wo_id = get_unnamed_command(cmd)
#     if cmd_wo_id in cmd_naming:
#         # TODO: immutable version, replace with "update" on the whole dictionary.
#         cmd_naming[cmd_wo_id] = add_ids(cmd_naming[cmd_wo_id], get_command_ids(cmd))
#     else:
#         output_cmds.append(cmd_wo_id)
#         cmd_naming[cmd_wo_id] = cmd

#     return cmd_naming, output_cmds

def please_instantiate_from_atom_ref(atom_ref: AtomRef, maybe_iid=None):

    d = {"atom": raet,
         "origin_uid": origin_uid(atom_ref)}
    if maybe_iid is not None:
        d["internal_ids"] = [maybe_iid]

    return PleaseInstantiate(d)
                              



def distinguish_cmd(cmd):
    # Returns the distinguishable representation of an object, without any other context known.
    #
    # Entity(ET.X, uid="abcd")[V.name] -> Entity(ET.X, uid="abcd")
    # ET.Machine[V.name] -> ET.Machine[InternalID(n+1)]
    # 
    # This is used to fill in the list of unique operations, which will be pointed at by other names later on.

    cmd_distinguish,names = cmd | match[
        distinguish_rules,
        (Any, not_implemented_error["Unknown cmd type"]),
    ] | collect

    return cmd_distinguish,names

def recombination(cmd1, cmd2):
    # When two objects have been identified as being the same, this is the procedure (or problem-aware point) for combining the two objects together.
    #
    # (ET.Machine[InternalID(1)], ET.Machine[InternalID(2)]) -> ET.Machine[InternalID(n+1)]
    # (Entity(ET.Machine, uid="abcd"), ET.Machine[InternalID(1)]) -> Entity(ET.Machine, uid="abcd")
    # (ET.X[InternalID(1)], ET.Y[InternalID(2)]) -> ERROR

    new_cmd = cmd | match[
        recombination_rules,
        (Any, not_implemented_error["Unknown cmd pair type"]),
    ] | collect

    return new_cmd
    

def add_command(cmd: PleaseCommandLevel1,
                name_mapping: Dict[WishID][Set[PleaseCommandLevel1Distinguishable]],
                cmd_names: Dict[PleaseCommandLevel1Distinguishable][Set[WishID]]):
    # Fully mutating style, but function signature at least allows for potential
    # of immutable implementation.

    # name_mapping - maps of names to cmds (cmds identified as values).
    #              - this uses a set of cmds, even though the final result from
    #                this function should only ever be sets of 1 cmd, i.e.
    #                unique. This is just for the intermediate values.
                
    # cmd_names - functions as both list of cmds and sets of names belonging to cmds.
    
    cmd_distinguish,names = distinguish_cmd(cmd)

    cmd_set.add(cmd_distinguish)

    names_needing_reconcilliation = []

    for name in names:
        name_mapping[name].add(cmd_distingiush)
        if len(name_mapping[name]) >= 2:
            names_needing_reconcilliation += [name]


    # Now we build up the connected subgraph of aliased name/cmd combinations.
    #
    # We could do this at the end of all commands instead. The advantage here is
    # that we have a very easy way to find the subgraph. Whereas doing it at the
    # end would require proper connected_components analysis.

    if len(names_needing_reconcilliation) != 0:
        from functools import reduce
        from operator import or_

        all_sets = [name_mapping[name] for name in names_needing_reconcilliation]
        all_cmds = reduce(_or, all_sets)

        resolve_connected_group(all_cmds)

    return name_mapping, cmd_names


def resolve_connected_group(group_cmds, name_mapping, cmd_names):
    all_names = reduce(_or, [cmd_names[cmd] for cmd in all_cmds])

    new_cmd = reduce(recombination, all_cmds)

    for cmd in all_cmds:
        del cmd_names[cmd]

    cmd_names[new_cmd] = all_names
    for name in all_names:
        name_mapping[name] = {new_cmd}